'''
GaussDiagram
============

.. image:: example_gauss_diagram_k10_93.png
   :alt: A Gauss diagram for the knot 10_93
   :align: center
   :scale: 50%

Class for creating and viewing Gauss diagrams.


API documentation
~~~~~~~~~~~~~~~~~
'''

from __future__ import print_function
import numpy as n


class GaussDiagram(object):
    '''
    Class for containing and manipulating Gauss diagrams.

    Parameters
    ----------
    representation : Another representation of a knot.
    '''

    def __init__(self, representation):
        from pyknotid.representations.gausscode import GaussCode

        gc = representation
        if not isinstance(gc, GaussCode):
            gc = GaussCode(representation)

        self._gauss_code = gc._gauss_code
        if len(self._gauss_code) == 0 or len(self._gauss_code) > 1:
            raise Exception('GaussDiagram can only be created from '
                            'single-component Gausscode')

        self._fig_ax = None
        self._fig_ax = self.plot()

    def plot(self, fig_ax=None):
        '''Plots the Gauss diagram using matplotlib. This is called
        automatically on __init__.

        Returns a tuple of the matplotlib figure and axis.
        '''
        import matplotlib.pyplot as plt
        if fig_ax is None:
            if self._fig_ax is not None:
                fig, ax = self._fig_ax
            else:
                fig, ax = plt.subplots()
        else:
            fig, ax = fig_ax
        ax.clear()

        c1 = plt.Circle((00, 00), radius=100., linewidth=2,
                        color='black',
                        fill=False, antialiased=True)

        gc = self._gauss_code[0]
        num_points = len(gc)

        angles = n.linspace(0, 2*n.pi, num_points+1)[:-1]

        other_angles = {}
        for i, angle in enumerate(angles):
            gc_entry = gc[i, 0]

            circ_pos = 100*n.array([n.cos(angle), n.sin(angle)])
            rad_pos = 1.2*circ_pos
            ax.annotate(str(gc_entry), xy=circ_pos, xytext=rad_pos)

            if gc_entry in other_angles:
                other_angle = other_angles.pop(gc_entry)
                other_pos = 100*n.array([n.cos(other_angle),
                                         n.sin(other_angle)])

                over = gc[i, 1]
                sign = gc[i, 2]
                colour = {-1: 'red', 1: 'blue'}[sign]
                if over > 0:
                    ax.annotate('', xy=other_pos, xytext=circ_pos,
                                arrowprops={'frac': 0.1, 'color': colour})
                else:
                    ax.annotate('', xytext=other_pos, xy=circ_pos,
                                arrowprops={'frac': 0.1, 'color': colour})
            else:
                other_angles[gc_entry] = angle

        ax.add_artist(c1)

        ax.plot(100*n.cos(angles), 100*n.sin(angles), 'o', markersize=8,
                color='black')

        ax.set_xlim(-140, 140)
        ax.set_ylim(-140, 140)
        ax.set_xticks([])
        ax.set_yticks([])
        fig.show()
        return fig, ax



from __future__ import print_function
from numpy import roll, zeros, cross, sqrt, sum, arcsin, isnan, pi


def higher_order_writhe_integral(points, order=(1, 3, 2, 4)):

    # The writhe contributions from each pair of segments
    # This is a matrix with side length the number of points in the curve
    contributions = zeros((len(points), len(points)))

    # This is just bookkeeping to ensure that the order of the writhe
    # comparison consists of the numbers 1 to 4
    assert len(set(order)) == len(order)
    assert set(order) == set((1, 2, 3, 4))
    order = [o - 1 for o in order]

    # First, we calculate the contribution to the writhe from every
    # pair of segments in the curve.
    for i1 in range(len(points) - 3):
        p1 = points[i]
        p2 = points[i1 + 1]
        for i2 in range(i1 + 2, len(points) - 1):
            p3 = points[i2]
            p4 = points[i2 + 1]

            # The algorithm here is just as from wikipedia

            r12 = p2 - p1
            r13 = p3 - p1
            r14 = p4 - p1
            r23 = p3 - p2
            r24 = p4 - p2
            r34 = p4 - p3

            n1 = cross(r13, r14)
            n1 /= sqrt(sum(n1**2))

            n2 = cross(r14, r24)
            n2 /= sqrt(sum(n2**2))

            n3 = cross(r24, r23)
            if any(abs(n3) > 0.):
                n3 /= sqrt(sum(n3**2))

            n4 = cross(r23, r13)
            if any(abs(n4) > 0.):
                n4 /= sqrt(sum(n4**2))

            # When the vectors are nearly the same, floating point
            # errors can apparently sometimes make the output a tiny
            # bit higher than 1. This code forces bounds on the
            # resulting values.
            t1, t2, t3, t4 = clip([n1.dot(n2),
                                   n2.dot(n3),
                                   n3.dot(n4),
                                   n4.dot(n1)],
                                  -1, 1)

            writhe_contribution = (arcsin(t1) +
                                   arcsin(t2) +
                                   arcsin(t3) +
                                   arcsin(t4))

            writhe_contribution *= sign(cross(r34, r12).dot(r13))
    
            contributions[i1, i2] = writhe_contribution
            contributions[i2, i1] = writhe_contribution

    # Now we can calculate the higher order writhe.

    writhe = 0.0

    for i1 in range(len(points) - 3):
        for i2 in range(i1 + 1, len(points) - 1):
            for i3 in range(i2 + 1, len(points) - 1):
                for i4 in range(i3 + 1, len(points) - 1):
                    indices = (i1, i2, i3, i4)

                    writhe += (contributions[indices[order[0]],
                                             indices[order[1]]] *
                               contributions[indices[order[2]],
                                             indices[order[3]]])
    print()

    return writhe / (2*pi)**2




import numpy as np
from os.path import dirname, exists, join, basename
from os import listdir
import glob
import gzip

from pyknot2.spacecurves import Knot

knots_folder = join(dirname(__file__), 'ideal_knots')

def _sort_func(filen):
    filen = filen.split('/')[-1]
    num_components = len(filen.split('#'))

    components = filen.split('#')[0].strip('_')

    values = components.split('_')

    crossings = int(values[0])
    number = int(values[1])

    return crossings + 1e-6 * number + 10000 * num_components

def available_ideal_knots():
    '''Returns a list of the knots whose conformations are available via
    :func:`ideal_knot`.
    '''

    filens = glob.glob(join(knots_folder, '*.txt.gz'))
    filens = [basename(filen)[:-7] for filen in filens]

    filens = sorted(filens, key=_sort_func)
    return filens

def ideal_knot(name):

    '''Returns a :class:`~pyknot2.spacecurves.knot.Knot` whose points form
    an approximation to the ideal (minimal ropelength) conformation of
    the given knot.

    These knots are the freely available tight knot conformations
    obtained by Jason Cantarella's Ridgerunner, see
    `http://www.jasoncantarella.com/wordpress/software/ridgerunner/`_
    for more information and the full data files.

    .. note:: Only prime_knots with up to 10 crossings are available,
              see :func:`available_ideal_knots`.

    Example::

        k = ideal_knot('6_3')

    Parameters
    ----------
    name : str
        The name of the knot in standard notation, e.g. 3_1, 4_1, ....
        Only knots with up to 10 crossings are available.

    '''

    filen = join(knots_folder, '{}.txt.gz'.format(name))
    if not exists(filen):
        raise ValueError(('Could not find filename {}.txt.gz for '
                          'the given knot name. Only knots with up '
                          'to 10 crossings are available.').format(name))

    points = []
    with gzip.open(filen, 'rb') as fileh:
        for line in fileh:
            line = line.decode('utf-8')
            numbers = line.split(' ')
            points.append([float(numbers[0]),
                           float(numbers[1]),
                           float(numbers[2])])

    return Knot(np.array(points))
                               
            




import numpy as n
cimport numpy as n
cimport cython

from pyknotid.utils import vprint

cdef long crude_modulus(long val, long modulo):
    if val < 0:
        return val + modulo
    return val
    

@cython.wraparound(False)
@cython.boundscheck(False)
cpdef vassiliev_degree_3(long [:, :] arrows):
    cdef long num_arrows = len(arrows)
    cdef long num_crossings = len(arrows) * 2

    cdef long a1s, a1e, a2s, a2e, a3s, a3e
    cdef long [:] arrow1, arrow2, arrow3
    cdef long i1, i2, i3
    cdef long sign1, sign2, sign3

    cdef set used_sets = set()
    cdef long representations_sum_1 = 0
    cdef long representations_sum_2 = 0
    cdef tuple ordered_indices
    for i1 in range(num_arrows):
        arrow1 = arrows[i1]
        a1s = arrow1[0]
        a1e = arrow1[1]
        sign1 = arrow1[2]

        a1e = crude_modulus(a1e - a1s, num_crossings)

        for i2 in range(num_arrows):
            arrow2 = arrows[i2]
            a2s = arrow2[0]
            a2e = arrow2[1]
            sign2 = arrow2[2]

            a2s = crude_modulus(a2s - a1s, num_crossings)
            a2e = crude_modulus(a2e - a1s, num_crossings)

            for i3 in range(num_arrows):
                arrow3 = arrows[i3]
                a3s = arrow3[0]
                a3e = arrow3[1]
                sign3 = arrow3[2]

                a3s = crude_modulus(a3s - a1s, num_crossings)
                a3e = crude_modulus(a3e - a1s, num_crossings)

                ordered_indices = tuple(sorted((i1, i2, i3)))
                if ordered_indices in used_sets:
                    continue

                if (a2s < a1e and a3e < a1e and a3e > a2s and
                    a3s > a1e and a2e > a3s):
                    representations_sum_1 += sign1 * sign2 * sign3
                    used_sets.add(ordered_indices)
                if (a2e < a1e and a3s < a1e and a3s > a2e and
                    a2s > a1e and a3e > a2s):
                    representations_sum_2 += sign1 * sign2 * sign3
                    used_sets.add(ordered_indices)

                    
    return representations_sum_1 / 2. + representations_sum_2

    




import numpy as np

from pyknot2.representations.gausscode import GaussCode

class ArrowDiagram(GaussCode):
    pass

class GaussDiagram(GaussCode):
    pass


class ReidemeisterMove(object):
    def __init__(self):
        self.before = []
        self.after = []

        self.builtin_crossings = {1}

    def get_subdiagrams(max_arrows=

class ReidemeisterOne(ReidemeisterMove):
    def __init__(self):

        self.before = [GaussCode('1+a,1-a')]

        self.after = []

        self.builtin_crossings = {1}


class ReidemeisterTwoA(ReidemeisterMove):
    def __init__(self):

        self.before = [GaussCode('1+a,2+c'),
                       GaussCode('2-c,1-a')]

        self.after = []

        self.builtin_crossings = {1, 2}
        
class ReidemeisterTwoB(ReidemeisterMove):
    def __init__(self):
        
        self.before = [GaussCode('1+a,2+c'),
                       GaussCode('1-a,2-c')]

        self.after = []

        self.builtin_crossings = {1, 2}

class ReidemeisterThreeA(ReidemeisterMove):
    def __init__(self):

        self.before = [GaussCode('1-a,2+c'),
                       GaussCode('1+a,3+c'),
                       GaussCode('2-c,3-c')]

        self.after = [GaussCode('2+c,1-a'),
                      GaussCode('3+c,1+a'),
                      GaussCode('3-c,2-c')]

        self.builtin_crossings = {1, 2, 3}

class ReidemeisterThreeB(ReidemeisterMove):
    def __init__(self):

        self.before = [GaussCode('1+c,2-a'),
                       GaussCode('1-c,3-a'),
                       GaussCode('2+a,3+a')]

        self.after = [GaussCode('2-a,1+c'),
                      GaussCode('3-a,1-c'),
                      GaussCode('3+a,2+a')]

        self.builtin_crossings = {1, 2, 3}




'''
This module provides functions for creating knotted or linked space
curves, for inspection or use as examples.

All knot creating functions are available in the pyknotid.make
namespace, although they are distributed amongst other modules as below.

'''
# from pyknotid.make import knot
# from pyknotid.make import randomwalks

# from pyknotid.make.ideal import available_ideal_knots, ideal_knot
from pyknotid.make.torus import torus_knot, torus_link
from pyknotid.make.named import *




from vispy.scene import Mesh

from vispy.visuals import MeshVisual

import numpy as np

from vispy.scene.visuals import create_visual_node

class MeshCollectionVisual(MeshVisual):
    def __init__(self, visuals, shading='smooth'):

        vertices = []
        indices = []
        vertex_colors = []

        for visual in visuals:
            if not isinstance(visual, MeshVisual):
                raise ValueError('Non-MeshVisual received')

            md = visual._meshdata
            vertices.append(md.get_vertices().copy())
            indices.append(md.get_faces().copy())
            vertex_colors.append(md.get_vertex_colors().copy())

        cum_lens = np.hstack([[0], np.cumsum(list(map(len, vertices)))[:-1]])

        for cur_len, inds in list(zip(cum_lens, indices)):
            inds += cur_len

        vertices = np.vstack(vertices)
        indices = np.vstack(indices)
        vertex_colors = np.vstack(vertex_colors)

        MeshVisual.__init__(self, vertices, indices,
                            vertex_colors=vertex_colors,
                            shading=shading)

            
MeshCollection = create_visual_node(MeshCollectionVisual)



'''pyknotid provides knot lookup by name or invariant values, using a
prebuilt database.

The knot database includes information about all knots with up to 15
crossings, with topological invariants following those indexed by the
`Knot Atlas <http://katlas.org/wiki/Main_Page>`__ and the `KnotInfo
Table of Knot Invariants <http://www.indiana.edu/~knotinfo/>`__, or
calculated by pyknotid using the Dowker-Thistlethwaite codes of the
knots.

Downloading the database
------------------------

The database must normally be downloaded separately, and is currently
approximately 230MB in size.

If you do not download the database, most of pyknotid will work
fine. Only the explicit knot identification by database lookup, or
direct database queries, are not available.

To download the knot database::

    from pyknotid.catalogue.getdb import download_database
    download_database()

After this has completed (it may take a few seconds), the database
functions should all work immediately.

For other database management functions, see :doc:`getdb`.

Lookup by name
--------------

Use :func:`pyknotid.catalogue.identify.get_knot`::

    from pyknotid.catalogue.identify import get_knot
    trefoil = get_knot('3_1')
    figure_eight = get_knot('4_1')

Lookup by invariants
--------------------
Use :func:`pyknotid.catalogue.identify.from_invariants`::

    from pyknotid.catalogue.identify import from_invariants

    from_invariants(determinant=5, max_crossings=9)
    # returns [<Knot 4_1>, <Knot 5_1>]

    import sympy as sym
    t = sym.var('t')
    from_invariants(alexander=1-t+t**2, max_crossings=9)
    # returns [<Knot 3_1>]

For a full list of lookup parameters, see :func:`~pyknotid.catalogue.identify.from_invariants`.

Exploring properties of knots
-----------------------------

You can view more properties of any knot returned by the database::

  from pyknotid.catalogue import get_knot, from_invariants

  k = get_knot('5_2')
  k.pretty_print()  # prints some information from the database:
                    #  Identifier: 5_2
                    #  Min crossings: 5
                    #  Fibered: False
                    #  Gauss code: -1, 5, -2, 1, -3, 4, -5, 2, -4, 3
                    #  Planar diagram: X_1425 X_3849 X_5,10,6,1 X_9,6,10,7 X_7283
                    #  DT code: 4 8 10 2 6
                    #  Determinant: 7
                    #  Signature: -2
                    #  Alexander: 2*t**2 - 3*t + 2
                    #  Jones: 1/q - 1/q**2 + 2/q**3 - 1/q**4 + q**(-5) - 1/q**6
                    #  HOMFLY: -a**6 + a**4*z**2 + a**4 + a**2*z**2 + a**2
                    #  Hyperbolic volume: 2.82812
                    #  Vassiliev order 2: 2
                    #  Vassiliev order 3: -3
                    #  Symmetry: reversible

Properties of the knot can also be accessed directly::

  k.determinant  # 7

For a full list of attributes available, see
:class:`pyknotid.catalogue.database.Knot`.

'''

from pyknotid.catalogue.identify import from_invariants, get_knot, first_from_invariants
from pyknotid.catalogue.getdb import download_database

db_version = 1



#!/usr/bin/env python2

from __future__ import print_function, division

import argparse
import sys
import numpy as np

from os import path

import pyknotid.visualise as pvis

from time import sleep

from os import path
sys.path.append(path.split(path.dirname(path.realpath(__file__)))[0])

from pyknotid.spacecurves.knot import Knot

from vispy import scene


def knot_from_file(filen):
    with open(filen, 'r') as fileh:
        text = fileh.read()

    return knot_from_text(text)
    

def knot_from_text(text):

    points = []
    for line in text.split('\n'):
        values = line.split(' ')
        if len(values) == 1 and len(values[0]) == 0:
            continue
        if len(values) != 3:
            raise ValueError('Invalid text passed.')
        points.append([float(v) for v in values])

    k = Knot(points, verbose=False)
    return k


def main():
    parser = argparse.ArgumentParser(
        description='Analyse the knot type of a curve in a data file')

    parser.add_argument('filename', help='The file to load')
    parser.add_argument('--output', help='The filepath to save an image to')

    parser.add_argument('--plot', dest='plot', action='store_true')
    parser.add_argument('--no-plot', dest='plot', action='store_false')
    parser.set_defaults(plot=True)

    parser.add_argument('--azimuth', help='Angle of camera around z axis, in degreesn',
                        default=30, type=float)
    parser.add_argument('--elevation', help='Height of camera above x-y plane, in degrees',
                        default=30, type=float)
    parser.add_argument('--scale-factor', help='Camera zoom',
                        type=float)

    args = parser.parse_args(sys.argv[1:])


    k = knot_from_file(args.filename)

    k.plot()

    vispy_canvas = pvis.vispy_canvas

    vispy_canvas.view.camera = scene.TurntableCamera(fov=30)
    camera = vispy_canvas.view.camera

    if args.azimuth:
        camera.azimuth = args.azimuth
    if args.elevation:
        camera.elevation = args.elevation
    if args.scale_factor is not None:
        camera.scale_factor = args.scale_factor

    if args.output:
        pvis.vispy_save_png(args.output)

    if args.plot:
        vispy_canvas.app.run()

    camera = pvis.vispy_canvas.view.camera

    print('final camera details:')
    print('  azimuth: {}'.format(camera.azimuth))
    print('  elevation: {}'.format(camera.elevation))
    print('  scale-factor: {}'.format(camera.scale_factor))

    print(('To reuse these, add the arguments `--azimuth {} '
           '--elevation {} --scale-factor {}').format(
               camera.azimuth, camera.elevation, camera.scale_factor))
    

if __name__ == "__main__":
    main()
    



'''
Space curve complexity
======================

Functions for evaluating the (mainly geometrical) complexity of space
curves.
'''

from __future__ import print_function

from . import Knot
from . import SpaceCurve
from pyknotid.spacecurves.rotation import (get_rotation_angles,
                                          rotate_to_top)
from pyknotid.utils import vprint
import numpy as n
import numpy as np

import sys


def writhe_and_crossing_number(points, number_of_samples=10,
                               verbose=True, remove_duplicate_angles=False,
                               include_closure=False,
                               **kwargs):
    '''
    Returns the writhe and crossing number of the given curve.

    Parameters
    ----------
    points : array-like
        An Nx3 array of points in the line.
    number_of_samples : int
        The number of projection directions to average over. These will
        be chosen to be roughly uniform on the sphere, as per
        :func:`get_rotation_angles`.
    remove_duplicate_angles : bool
        Whether to remove duplicate angles (i.e. points exactly opposite
        one another on the sphere), adjusting the average appropriately.
        This argument is currently experimental and defaults to False;
        do *not* trust it to remove points correctly.
    **kwargs :
        These are passed to the raw_crossings method of Knot classes
        used internally.
    '''

    angles = get_rotation_angles(number_of_samples)
    if remove_duplicate_angles:
        angles = angles[:int(len(angles) / 2.)]

    crossing_numbers = []
    writhes = []

    for theta, phi in angles:
        k = SpaceCurve(points, verbose=False)
        k._apply_matrix(rotate_to_top(theta, phi))
        crossings = k.raw_crossings(include_closure=include_closure, **kwargs)
        crossing_numbers.append(len(crossings) / 2)
        writhes.append(n.sum(crossings[:, 3]) / 2. if len(crossings) else 0)

    return n.average(crossing_numbers), n.average(writhes)


# def writhe_integral(points, closed=False):

#     ps = points

#     dps = np.roll(points, -1, axis=0) - points

#     writhe = 0.0

#     points2 = points[:-1] if not closed else points

#     for i1, p1 in enumerate(points2):
#         for i2, p2 in enumerate(points2[i1+1:]):
#             i2 += i1 + 1

#             diff = p2 - p1
#             cross = np.cross(dps[i1], dps[i2])
#             writhe += diff.dot(cross) / np.sqrt(np.sum(diff**2))**3

#     return 2 * writhe / (4*np.pi)


def writhe_integral(points, closed=False):

    ps = points

    dps = np.roll(points, -1, axis=0) - points

    writhe = 0.0

    points2 = points[:-1] if not closed else points

    for i1, p1 in enumerate(points[:-3]):
        p2 = points[i1 + 1]
        for i2, p3 in enumerate(points[i1+2:-1]):
            # p2 = p2 + 0.000001
            i2 += i1 + 2

            p4 = points[i2 + 1]

            r12 = p2 - p1
            r13 = p3 - p1
            r14 = p4 - p1
            r23 = p3 - p2
            r24 = p4 - p2
            r34 = p4 - p3

            n1 = np.cross(r13, r14)
            n1 /= np.sqrt(np.sum(n1**2))

            n2 = np.cross(r14, r24)
            n2 /= np.sqrt(np.sum(n2**2))

            n3 = np.cross(r24, r23)
            if np.any(np.abs(n3) > 0.):
                n3 /= np.sqrt(np.sum(n3**2))

            n4 = np.cross(r23, r13)
            if np.any(np.abs(n4) > 0.):
                n4 /= np.sqrt(np.sum(n4**2))

            if np.any(np.isnan(n3)):
                print('!!! nan')
                print(i1, i2)
                print(p1, p2, p3, p4)
                print('nan', r23, r13, np.cross(r23, r13))

            t1, t2, t3, t4 = np.clip([n1.dot(n2),
                                      n2.dot(n3),
                                      n3.dot(n4),
                                      n4.dot(n1)],
                                     -1, 1)

            writhe_contribution = (np.arcsin(t1) +
                                   np.arcsin(t2) +
                                   np.arcsin(t3) +
                                   np.arcsin(t4))

            writhe_contribution *= np.sign(np.cross(r34, r12).dot(r13))

            if np.isnan(writhe_contribution):
                print()
                print('nan')
                print(i1, i2, n1, n2, n3, n4)

            writhe += writhe_contribution


    return 2 * writhe / (4*np.pi)

def higher_order_writhe_integral(points, order=(1, 3, 2, 4), try_cython=True):

    ps = points
    dps = np.roll(points, -1, axis=0) - points

    contributions = np.zeros((len(points[:-1]), len(points[:-1])))
    contributions = np.zeros((len(points), len(points)))

    assert len(set(order)) == len(order)
    assert set(order) == set((1, 2, 3, 4))

    print('Calculating writhe contributions')

    order = [o - 1 for o in order]
    for i1, p1 in enumerate(points[:-3]):
        print('\ri = {} / {}'.format(i1, len(points) - 4), end='')
        sys.stdout.flush()
        p2 = points[i1 + 1]
        for i2, p3 in enumerate(points[i1+2:-1]):
            i2 += i1 + 2

            p4 = points[i2 + 1]

            r12 = p2 - p1
            r13 = p3 - p1
            r14 = p4 - p1
            r23 = p3 - p2
            r24 = p4 - p2
            r34 = p4 - p3

            n1 = np.cross(r13, r14)
            n1 /= np.sqrt(np.sum(n1**2))

            n2 = np.cross(r14, r24)
            n2 /= np.sqrt(np.sum(n2**2))

            n3 = np.cross(r24, r23)
            if np.any(np.abs(n3) > 0.):
                n3 /= np.sqrt(np.sum(n3**2))

            n4 = np.cross(r23, r13)
            if np.any(np.abs(n4) > 0.):
                n4 /= np.sqrt(np.sum(n4**2))

            if np.any(np.isnan(n1)):
                print('!!! nan')
                print(i1, i2)
                print(p1, p2, p3, p4)
                print('nan', r23, r13, np.cross(r23, r13))

            # When the vectors are nearly the same, floating point
            # errors can sometimes make the output a tiny bit higher
            # than 1
            t1, t2, t3, t4 = np.clip([n1.dot(n2),
                                      n2.dot(n3),
                                      n3.dot(n4),
                                      n4.dot(n1)],
                                     -1, 1)

            writhe_contribution = (np.arcsin(t1) +
                                   np.arcsin(t2) +
                                   np.arcsin(t3) +
                                   np.arcsin(t4))

            if np.isnan(writhe_contribution):
                print()
                print('nan!')
                print(i1, i2, n1, n2, n3, n4, writhe_contribution)
                print(n1.dot(n2) > 1, n2.dot(n3) > 1, n3.dot(n4) > 1, n4.dot(n1) > 1)
                print(n1.dot(n2), np.arcsin(n1.dot(n2)))
                print(n2.dot(n3), np.arcsin(n2.dot(n3)))
                print(n3.dot(n4), np.arcsin(n3.dot(n4)))
                print(n4.dot(n1), np.arcsin(n4.dot(n1)))

            writhe_contribution *= np.sign(np.cross(r34, r12).dot(r13))
    
            contributions[i1, i2] = writhe_contribution
            contributions[i2, i1] = writhe_contribution

    print()
    print('Calculating higher order writhe')

    how_func = _higher_order_writhe

    if try_cython:
        try:
            from pyknotid.spacecurves.ccomplexity import cython_higher_order_writhe 
            order = np.array(order)
            how_func = cython_higher_order_writhe
        except ImportError:
            print('Failed to import ccomplexity, using pure python instead')
    
    writhe = how_func(points, contributions, order)

    return writhe / (2*np.pi)**2

def _higher_order_writhe(points, contributions, order):
    writhe = 0.0

    for i1, p1s in enumerate(points[:-3]):
        print('\ri1', i1, len(points) - 4, end='')
        sys.stdout.flush()
        for i2, p2s in enumerate(points[i1+1:-1]):
            i2 += i1 + 1
            for i3, p3s in enumerate(points[i2+1:-1]):
                i3 += i2 + 1
                for i4, p4s in enumerate(points[i3+1:-1]):
                    i4 += i3 + 1

                    indices = (i1, i2, i3, i4)
                    # print('indices', indices)
                    # print('shape', contributions.shape)
                    writhe += (contributions[indices[order[0]],
                                             indices[order[1]]] *
                               contributions[indices[order[2]],
                                             indices[order[3]]])
    print()

    return writhe


def writhing_matrix(points, order=(1, 3, 2, 4)):

    ps = points
    dps = np.roll(points, -1, axis=0) - points

    contributions = np.zeros((len(points[:-1]), len(points[:-1])))
    contributions = np.zeros((len(points), len(points)))

    assert len(set(order)) == len(order)
    assert set(order) == set((1, 2, 3, 4))

    print('Calculating writhe contributions')

    order = [o - 1 for o in order]

    for i1, p1 in enumerate(points[:-3]):
        if i1 % 50 == 0:
            print('\ri = {} / {}'.format(i1, len(points) - 4), end='')
        p2 = points[i1 + 1]
        for i2, p3 in enumerate(points[i1+2:-1]):
            # p2 = p2 + 0.000001
            i2 += i1 + 2

            p4 = points[i2 + 1]

            r12 = p2 - p1
            r13 = p3 - p1
            r14 = p4 - p1
            r23 = p3 - p2
            r24 = p4 - p2
            r34 = p4 - p3

            n1 = np.cross(r13, r14)
            n1 /= np.sqrt(np.sum(n1**2))

            n2 = np.cross(r14, r24)
            n2 /= np.sqrt(np.sum(n2**2))

            n3 = np.cross(r24, r23)
            if np.any(np.abs(n3) > 0.):
                n3 /= np.sqrt(np.sum(n3**2))

            n4 = np.cross(r23, r13)
            if np.any(np.abs(n4) > 0.):
                n4 /= np.sqrt(np.sum(n4**2))

            if np.any(np.isnan(n1)):
                print('!!! nan')
                print(i1, i2)
                print(p1, p2, p3, p4)
                print('nan', r23, r13, np.cross(r23, r13))

            # When the vectors are nearly the same, floating point
            # errors can sometimes make the output a tiny bit higher
            # than 1
            t1, t2, t3, t4 = np.clip([n1.dot(n2),
                                      n2.dot(n3),
                                      n3.dot(n4),
                                      n4.dot(n1)],
                                     -1, 1)

            writhe_contribution = (np.arcsin(t1) +
                                   np.arcsin(t2) +
                                   np.arcsin(t3) +
                                   np.arcsin(t4))

            if np.isnan(writhe_contribution):
                print()
                print('nan!')
                print(i1, i2, n1, n2, n3, n4, writhe_contribution)
                print(n1.dot(n2) > 1, n2.dot(n3) > 1, n3.dot(n4) > 1, n4.dot(n1) > 1)
                print(n1.dot(n2), np.arcsin(n1.dot(n2)))
                print(n2.dot(n3), np.arcsin(n2.dot(n3)))
                print(n3.dot(n4), np.arcsin(n3.dot(n4)))
                print(n4.dot(n1), np.arcsin(n4.dot(n1)))

            writhe_contribution *= np.sign(np.cross(r34, r12).dot(r13))
    
            contributions[i1, i2] = writhe_contribution
            contributions[i2, i1] = writhe_contribution

    print()

    return contributions


def second_order_writhes(points, try_cython=True, basepoint=True):

    contributions = np.zeros((len(points[:-1]), len(points[:-1])))
    contributions = np.zeros((len(points), len(points)))

    print('Calculating writhe contributions')

    for i1, p1 in enumerate(points[:-3]):
        print('\ri = {} / {}'.format(i1, len(points) - 4), end='')
        sys.stdout.flush()
        p2 = points[i1 + 1]
        for i2, p3 in enumerate(points[i1+2:-1]):
            i2 += i1 + 2

            p4 = points[i2 + 1]

            r12 = p2 - p1
            r13 = p3 - p1
            r14 = p4 - p1
            r23 = p3 - p2
            r24 = p4 - p2
            r34 = p4 - p3

            n1 = np.cross(r13, r14)
            n1 /= np.sqrt(np.sum(n1**2))

            n2 = np.cross(r14, r24)
            n2 /= np.sqrt(np.sum(n2**2))

            n3 = np.cross(r24, r23)
            if np.any(np.abs(n3) > 0.):
                n3 /= np.sqrt(np.sum(n3**2))

            n4 = np.cross(r23, r13)
            if np.any(np.abs(n4) > 0.):
                n4 /= np.sqrt(np.sum(n4**2))

            if np.any(np.isnan(n1)):
                print('!!! nan')
                print(i1, i2)
                print(p1, p2, p3, p4)
                print('nan', r23, r13, np.cross(r23, r13))

            # When the vectors are nearly the same, floating point
            # errors can sometimes make the output a tiny bit higher
            # than 1
            t1, t2, t3, t4 = np.clip([n1.dot(n2),
                                      n2.dot(n3),
                                      n3.dot(n4),
                                      n4.dot(n1)],
                                     -1, 1)

            writhe_contribution = (np.arcsin(t1) +
                                   np.arcsin(t2) +
                                   np.arcsin(t3) +
                                   np.arcsin(t4))

            if np.isnan(writhe_contribution):
                print()
                print('nan!')
                print(i1, i2, n1, n2, n3, n4, writhe_contribution)
                print(n1.dot(n2) > 1, n2.dot(n3) > 1, n3.dot(n4) > 1, n4.dot(n1) > 1)
                print(n1.dot(n2), np.arcsin(n1.dot(n2)))
                print(n2.dot(n3), np.arcsin(n2.dot(n3)))
                print(n3.dot(n4), np.arcsin(n3.dot(n4)))
                print(n4.dot(n1), np.arcsin(n4.dot(n1)))

            writhe_contribution *= np.sign(np.cross(r34, r12).dot(r13))
    
            contributions[i1, i2] = writhe_contribution
            contributions[i2, i1] = writhe_contribution

    print()
    print('Calculating higher order writhe')


    if basepoint:
        from pyknotid.spacecurves.ccomplexity import cython_second_order_writhes
    else:
        from pyknotid.spacecurves.ccomplexity import cython_second_order_writhes_no_basepoint as cython_second_order_writhes

    return cython_second_order_writhes(points, contributions)


def second_order_twist(points, z):
    
    contributions = np.zeros((len(points[:-1]), len(points[:-1])))
    contributions = np.zeros((len(points), len(points)))

    contributions = np.zeros(len(points))

    print('Calculating writhe contributions')

    for i1, p1 in enumerate(points[:-3]):
        print('\ri = {} / {}'.format(i1, len(points) - 4), end='')
        sys.stdout.flush()
        p2 = points[i1 + 1]

        p3 = z
        p4 = z

        print('p3 is', p3)
        print('p4 is', p4)

        r12 = p2 - p1
        r13 = p3 - p1
        r14 = p4 - p1
        r23 = p3 - p2
        r24 = p4 - p2
        # r34 = p4 - p3

        # n1 = np.cross(r13, r14)
        # n1 /= np.sqrt(np.sum(n1**2))
        n1 = np.array([0., 0, 0])

        n2 = np.cross(r14, r24)
        n2 /= np.sqrt(np.sum(n2**2))

        # n3 = np.cross(r24, r23)
        # if np.any(np.abs(n3) > 0.):
        #     n3 /= np.sqrt(np.sum(n3**2))
        n3 = np.array([0., 0, 0])

        n4 = np.cross(r23, r13)
        if np.any(np.abs(n4) > 0.):
            n4 /= np.sqrt(np.sum(n4**2))

        if np.any(np.isnan(n1)):
            print('!!! nan')
            # print(i1, i2)
            print(p1, p2, p3, p4)
            print('nan', r23, r13, np.cross(r23, r13))
            return

        # When the vectors are nearly the same, floating point
        # errors can sometimes make the output a tiny bit higher
        # than 1
        t1, t2, t3, t4 = np.clip([n1.dot(n2),
                                  n2.dot(n3),
                                  n3.dot(n4),
                                  n4.dot(n1)],
                                 -1, 1)

        writhe_contribution = (np.arcsin(t1) +
                                np.arcsin(t2) +
                                np.arcsin(t3) +
                                np.arcsin(t4))

        if np.isnan(writhe_contribution):
            print()
            print('nan!')
            print(i1, n1, n2, n3, n4, writhe_contribution)
            print(n1.dot(n2) > 1, n2.dot(n3) > 1, n3.dot(n4) > 1, n4.dot(n1) > 1)
            print(n1.dot(n2), np.arcsin(n1.dot(n2)))
            print(n2.dot(n3), np.arcsin(n2.dot(n3)))
            print(n3.dot(n4), np.arcsin(n3.dot(n4)))
            print(n4.dot(n1), np.arcsin(n4.dot(n1)))

        # writhe_contribution *= np.sign(np.cross(r34, r12).dot(r13))

        contributions[i1] = writhe_contribution

    return contributions




'''
Functions for making knots, returning a set of points.
'''

import numpy as np

from pyknot2.make import knot

def trefoil(slip_length=30):

    data = knot.trefoil() * 5

    return np.vstack((data[:-2], data[-slip_length:-2][::-1] + 1.5))

    



'''Visualise
=========

Functions for plotting knots, supporting different toolkits and types
of plot.

pyknotid primarily supports `Vispy <http://vispy.org/>`__ as the
plotting mechanism. `Mayavi
<http://docs.enthought.com/mayavi/mayavi/>`__ is semi-supported but
may not always work.

Many of these functions can be called in a more convenient way via
methods of the :doc:`space curve classes <spacecurves/index>`
(e.g. :class:`~pyknotid.spacecurves.knot.Knot`).

API documentation
-----------------

'''

from __future__ import division

import vispy
# vispy.use('PyQt5')

import numpy as n
import numpy as np
from colorsys import hsv_to_rgb
from pyknotid.utils import ensure_shape_tuple, vprint
import random
from colorsys import hsv_to_rgb

vispy_canvas = None

def plot_vispy_cube():
    ensure_vispy_canvas()
    clear_vispy_canvas()
    canvas = vispy_canvas

    from vispy import app, scene, color

    c = scene.visuals.Cube((5, 2, 10),
                           color='blue',
                           edge_color='red')
    canvas.view.add(c)
    canvas.view.camera = scene.ArcballCamera(fov=30, distance=20)
    canvas.show()

def plot_line(points, mode='auto', clf=True, **kwargs):
    '''
    Plots the given line, using the toolkit given by mode.

    kwargs are passed to the toolkit specific function, except for:

    Parameters
    ----------
    points : ndarray
        The nx3 array to plot.
    mode : str
        The toolkit to draw with. Defaults to 'auto', which will
        automatically pick the first available toolkit from
        ['mayavi', 'matplotlib', 'vispy'], or raise an exception
        if none can be imported.
    clf : bool
        Whether the existing figure should be cleared
        before drawing the new one.
    '''
    if mode == 'auto':
        try:
            import vispy
            mode = 'vispy'
        except ImportError:
            pass
    if mode == 'auto':
        try:
            import mayavi.mlab as may
            mode = 'mayavi'
        except ImportError:
            pass
    if mode == 'auto':
        try:
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D
            fig = plt.figure()
            ax = fig.add_subplot(111, projection='3d')
            mode = 'matplotlib'
        except (ImportError, ValueError):
            pass
    if mode == 'auto':
        raise ImportError('Couldn\'t import any of mayavi, vispy, '
                          'or matplotlib\'s 3d axes.')
            
    if mode == 'mayavi':
        plot_line_mayavi(points, clf=clf, **kwargs)
    elif mode == 'vispy':
        plot_line_vispy(points, clf=clf, **kwargs)
    elif mode == 'matplotlib':
        plot_line_matplotlib(points, clf=clf, **kwargs)
    else:
        raise Exception('invalid toolkit/mode')

    
def plot_line_mayavi(points, clf=True, tube_radius=1., colormap='hsv',
                     closed=True,
                     zero_centroid=False,
                     mus=None,
                     **kwargs):
    import mayavi.mlab as may
    if clf:
        may.clf()
    if mus is None:
        mus = n.linspace(0, 1, len(points))
    may.plot3d(points[:, 0], points[:, 1], points[:, 2], mus,
               colormap=colormap, tube_radius=tube_radius, **kwargs)

def plot_line_matplotlib(points, **kwargs):
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = fig.gca(projection='3d')
    ax.plot(points[:, 0], points[:, 1], points[:, 2])
    fig.show()

def ensure_vispy_canvas():
    global vispy_canvas
    if vispy_canvas is None:
        from vispy import app, scene
        canvas = scene.SceneCanvas(keys='interactive', bgcolor='white')
        try:
            canvas.unfreeze()
        except AttributeError:  # depends on Vispy version
            pass
        canvas.view = canvas.central_widget.add_view()
        vispy_canvas = canvas
    # if not vispy_canvas.central_widget.children:
    #     vispy_canvas.view = vispy_canvas.central_widget.add_view()
        

def clear_vispy_canvas():
    global vispy_canvas
    if vispy_canvas is None:
        return
    try:
        vispy_canvas.unfreeze()
    except AttributeError:  # depends on Vispy version
        pass
    vispy_canvas.central_widget.remove_widget(vispy_canvas.view)
    vispy_canvas.view = vispy_canvas.central_widget.add_view()

def vispy_rotate(elevation=0, max_angle=360):
    global vispy_canvas
    cam = vispy_canvas.view.camera
    from vispy.scene import TurntableCamera
    from time import sleep
    vispy_canvas.view.camera = TurntableCamera(
        fov=cam.fov, scale_factor=cam.scale_factor,
        azimuth=0, elevation=elevation)
    try:
        for i in range(max_angle):
            vispy_canvas.view.camera.azimuth = i
            vispy_canvas.update()
            vispy_canvas.events.draw()
            vispy_canvas.swap_buffers()
            sleep(1 / 60.)
    except KeyboardInterrupt:
        pass
    vispy_canvas.view.camera = cam


def plot_line_vispy(points, clf=True, tube_radius=1.,
                    colour=None, zero_centroid=True,
                    closed=False, mus=None,
                    cmap=None,
                    tube_points=8, **kwargs):
    # Add an extra point to fix tube drawing bug
    last_tangent = points[-1] - points[-2]
    points = n.vstack([points, points[-1] + 0.0001 * last_tangent])
    
    ensure_vispy_canvas()
    if clf:
        clear_vispy_canvas()
    canvas = vispy_canvas
    from vispy import app, scene, color

    if isinstance(cmap, str):
        from matplotlib.cm import get_cmap
        mpl_cmap = get_cmap(cmap)
        cmap = lambda v: n.array(mpl_cmap(v))
    cmap = cmap or (lambda c: hsv_to_rgb(c, 1, 1))

    if colour is None:
        colours = n.linspace(0, 1, len(points))
        colours = n.array([cmap(c) for c in colours])
    else:
        colours = color.ColorArray(colour)

    if mus is not None:
        colours = n.array([hsv_to_rgb(c, 1, 1) for c in mus])
            

    l = scene.visuals.Tube(points, color=colours,
                           shading='smooth',
                           radius=tube_radius,
                           tube_points=tube_points,
                           closed=closed)
    
    canvas.view.add(l)
    # canvas.view.camera = 'arcball'
    canvas.view.camera = scene.ArcballCamera(fov=30, distance=7.5*n.max(
        n.abs(points)))
    #canvas.view.camera = scene.TurntableCamera(fov=30)
    if zero_centroid:
        l.transform = scene.transforms.MatrixTransform()
        # l.transform = scene.transforms.AffineTransform()
        l.transform.translate(-1*n.average(points, axis=0))

    canvas.show()
    # import ipdb
    # ipdb.set_trace()
    return canvas
    
def plot_lines_vispy(lines, clf=True, tube_radius=1.,
                     colours=None, zero_centroid=True, tube_points=8,
                     closed=False,
                     **kwargs):
    ensure_vispy_canvas()
    if clf:
        clear_vispy_canvas()
    canvas = vispy_canvas
    from vispy import app, scene, color

    if not isinstance(tube_radius, list):
        tube_radius = [tube_radius for _ in range(len(lines))]

    if colours is None:
        colours = ['purple' for line in lines]

    tubes = []
    for colour, points, radius in zip(colours, lines, tube_radius):

        l = scene.visuals.Tube(points, color=colour,
                               shading='smooth',
                               radius=radius,
                               closed=closed,
                               tube_points=tube_points)
        tubes.append(l)
    
    from .visualcollection import MeshCollection
    collection = MeshCollection(tubes)
    canvas.view.add(collection)
    canvas.view.camera = 'arcball'
    canvas.view.camera.fov = 30
    # canvas.view.camera = scene.TurntableCamera(
    #     fov=90, up='z', distance=1.2*n.max(n.max(
    #         points, axis=0)))

    if zero_centroid:
        l.transform = scene.transforms.MatrixTransform()
        # l.transform = scene.transforms.AffineTransform()
        l.transform.translate(-1*n.average(points, axis=0))

    canvas.show()
    return canvas
    

def plot_projection(points, crossings=None, mark_start=False,
                    fig_ax=None, show=True, mark_points=False):
    '''
    Plot the 2d projection of the given points, with optional
    markers for where the crossings are.

    Parameters
    ----------
    points : array-like
        The nxm array of points in the line, with m >= 2.
    crossings : array-like or None
        The nx2 array of crossing positions. If None, crossings
        are not plotted. Defaults to None.
    '''
    import matplotlib.pyplot as plt

    if fig_ax is not None:
        fig, ax = fig_ax
    else:
        fig, ax = plt.subplots()
    ax.plot(points[:, 0], points[:, 1], 'o-' if mark_points else '-')
    ax.set_xticks([])
    ax.set_yticks([])

    xmin, ymin = n.min(points[:, :2], axis=0)
    xmax, ymax = n.max(points[:, :2], axis=0)
    dx = (xmax - xmin) / 10.
    dy = (ymax - ymin) / 10.

    ax.set_xlim(xmin - dx, xmax + dx)
    ax.set_ylim(ymin - dy, ymax + dy)

    if mark_start:
        ax.plot([points[0, 0]], [points[0, 1]], color='blue',
                marker='o')
    
    if crossings is not None and len(crossings):
        crossings = n.array(crossings)
        ax.plot(crossings[:, 0], crossings[:, 1], 'ro', alpha=0.5)
    if show:
        fig.show()

    return fig, ax


def plot_shell(func, points, mode='auto', **kwargs):
    if mode =='auto':
        try:
            import vispy
            mode = 'vispy'
        except ImportError:
            pass
    if mode == 'auto':
        try:
            import mayavi.mlab as may
            mode = 'mayavi'
        except ImportError:
            pass

    if mode == 'mayavi':
        plot_shell_mayavi(func, points, **kwargs)
    elif mode == 'vispy':
        plot_shell_vispy(func, points, **kwargs)
    else:
        raise ValueError('invalid toolkit/mode')

def plot_shell_mayavi(func,
                      points,
                      number_of_samples=10,
                      zero_centroid=False,
                      sphere_radius_factor=2.,
                      opacity=0.3, **kwargs):
    positions, values = func(
        number_of_samples, zero_centroid=zero_centroid)

    thetas = n.arcsin(n.linspace(-1, 1, 100)) + n.pi / 2.
    phis = n.linspace(0, 2 * n.pi, 157)

    thetas, phis = n.meshgrid(thetas, phis)

    r = sphere_radius_factor * n.max(points)
    zs = r * n.cos(thetas)
    xs = r * n.sin(thetas) * n.cos(phis)
    ys = r * n.sin(thetas) * n.sin(phis)

    import mayavi.mlab as may

    may.mesh(xs, ys, zs, scalars=values, opacity=opacity, **kwargs)

def plot_sphere_shell_vispy(func, rows=100, cols=100,
                            radius=1.,
                            opacity=1.,
                            translation=(0., 0., 0.),
                            method='latitude',
                            edge_color=None,
                            cmap='hsv',
                            smooth=0,
                            cutoff=0.4,
                            cutoff_max=0.8,
                            transparent_side=True,
                            **kwargs):
    '''func must be a function of sphere angles theta, phi'''
    
    from vispy.scene import Sphere, ArcballCamera

    s = Sphere(rows=rows, cols=cols, method=method,
               edge_color=edge_color,
               radius=radius)
    mesh = s.mesh
    md = mesh._meshdata
    vertices = md.get_vertices()
    md.set_vertices(vertices + n.array(translation))

    values = n.zeros(len(vertices))
    opacities = n.ones(len(vertices))

    print('pre')
    for i, vertex in enumerate(vertices):
        if i % 10 == 0:
            vprint('\ri = {} / {}'.format(i, len(vertices)), newline=False)
        vertex = vertex / n.sqrt(n.sum(vertex*vertex))
        theta = n.arccos(vertex[2])
        phi = n.arctan2(vertex[1], vertex[0])

        if n.isnan(theta):
            theta = 0.0
        values[i] = func(theta, phi)
        distance = vertex[1] - cutoff
        distance_frac = distance / (cutoff_max - cutoff)
        opacity = max(0, 1. - distance_frac)
        opacities[i] = 1. if vertex[1] < cutoff else opacity
    vprint()

    colours = n.zeros((len(values), 4))
    max_val = n.max(values)
    min_val = n.min(values)
    unique_values = n.unique(colours)
    # max_val += (1. + 1./len(unique_values))*(max_val - min_val)
    diff = (max_val - min_val)

    import matplotlib.pyplot as plt
    cm = plt.get_cmap(cmap)
    for i in range(len(colours)):
        colours[i] = cm(((values[i] - min_val) / diff))

    colours[:, -1] = opacity

    faces = md.get_faces()
    for si in range(smooth):
        new_colours = [[n.array(row) for _ in range(3)]
                       for row in colours]
        for i, face in enumerate(faces):
            new_colours[face[0]].append(colours[face[1]])
            new_colours[face[0]].append(colours[face[2]])
            new_colours[face[1]].append(colours[face[0]])
            new_colours[face[1]].append(colours[face[2]])
            new_colours[face[2]].append(colours[face[0]])
            new_colours[face[2]].append(colours[face[1]])

        new_colours = n.array([n.average(cs, axis=0) for cs in new_colours])

        colours = new_colours

    colours = [(c[0], c[1], c[2], o) for c, o in zip(colours, opacities)]

    md.set_vertex_colors(colours)

    ensure_vispy_canvas()
    vispy_canvas.view.camera = ArcballCamera(fov=30)
    vispy_canvas.view.add(s)
    vispy_canvas.show()

    return colours
                            

def plot_shell_vispy(func,
                     points,
                     number_of_samples=10,
                     radius=None,
                     zero_centroid=False,
                     sphere_radius_factor=2.,
                     opacity=0.5,
                     cmap='hsv',
                     **kwargs):
    '''func must be a function returning values at angles and points,
    like OpenKnot._alexander_map_values. 
    '''
    
    positions, values = func(
        number_of_samples, radius=radius,
        zero_centroid=zero_centroid)
    
    thetas = n.arcsin(n.linspace(-1, 1, 100)) + n.pi/2.
    phis = n.linspace(0, 2*n.pi, 157)
    
    thetas, phis = n.meshgrid(thetas, phis)
    
    r = sphere_radius_factor*n.max(points)
    zs = r*n.cos(thetas)
    xs = r*n.sin(thetas)*n.cos(phis)
    ys = r*n.sin(thetas)*n.sin(phis)
    
    colours = n.zeros((values.shape[0], values.shape[1], 4))
    max_val = n.max(values)
    min_val = n.min(values)
    unique_values = n.unique(colours)
    max_val += (1. + 1./len(unique_values))*(max_val - min_val)
    diff = (max_val - min_val)

    import matplotlib.pyplot as plt
    cm = plt.get_cmap(cmap)
    for i in range(colours.shape[0]):
        for j in range(colours.shape[1]):
            colours[i, j] = cm(((values[i, j] - min_val) / diff))

    colours[:, :, -1] = opacity

    from vispy.scene import GridMesh
    from pyknotid.visualise import vispy_canvas
    mesh = GridMesh(xs, ys, zs, colors=colours)
    vispy_canvas.view.add(mesh)


def plot_cell(lines, mode='auto', **kwargs):
    if mode == 'auto':
        try:
            import vispy
            mode = 'vispy'
        except ImportError:
            pass
    if mode == 'auto':
        try:
            import mayavi.mlab as may
            mode = 'mayavi'
        except ImportError:
            pass

    if mode == 'mayavi':
        plot_cell_mayavi(lines, **kwargs)
    elif mode == 'vispy':
        plot_cell_vispy(lines, zero_centroid=False, **kwargs)
    else:
        raise ValueError('invalid toolkit/mode')

def plot_cell_mayavi(lines, boundary=None, clf=True, smooth=True,
                     min_seg_length=5, **kwargs):
    import mayavi.mlab as may
    may.clf()

    hues = n.linspace(0, 1, len(lines) + 1)[:-1]
    colours = [hsv_to_rgb(hue, 1, 1) for hue in hues]
    random.shuffle(colours)
    i = 0
    for (line, colour) in zip(lines, colours):
        vprint('Plotting line {} / {}\r'.format(i, len(lines)-1),
               False)
        i += 1
        for segment in line:
            if len(segment) < min_seg_length:
                continue
            plot_line(segment, mode='mayavi',
                      clf=False, color=colour, **kwargs)
    
    if boundary is not None:
        draw_bounding_box_mayavi(boundary, **kwargs)

def draw_bounding_box_mayavi(shape, colour=(0, 0, 0), tube_radius=1, markz=False):
    if shape is not None:
        if isinstance(shape, (float, int)):
            shape = ensure_shape_tuple(shape)
        if len(shape) == 3:
            shape = (0, shape[0], 0, shape[1], 0, shape[2])
    import mayavi.mlab as may

    xmin, xmax, ymin, ymax, zmin, zmax = shape
    ls = []
    ls.append(n.array([[xmax, ymax, zmin],[xmax, ymax, zmax]]))
    ls.append(n.array([[xmax, ymin, zmin],[xmax, ymin, zmax]]))
    ls.append(n.array([[xmin, ymax, zmin],[xmin, ymax, zmax]]))
    ls.append(n.array([[xmin, ymin, zmin],[xmin, ymin, zmax]]))
    ls.append(n.array([[xmin, ymax, zmax],[xmax, ymax, zmax]]))
    ls.append(n.array([[xmin, ymin, zmax],[xmax, ymin, zmax]]))
    ls.append(n.array([[xmin, ymax, zmin],[xmax, ymax, zmin]]))
    ls.append(n.array([[xmin, ymin, zmin],[xmax, ymin, zmin]]))
    ls.append(n.array([[xmax, ymin, zmax],[xmax, ymax, zmax]]))
    ls.append(n.array([[xmin, ymin, zmax],[xmin, ymax, zmax]]))
    ls.append(n.array([[xmax, ymin, zmin],[xmax, ymax, zmin]]))
    ls.append(n.array([[xmin, ymin, zmin],[xmin, ymax, zmin]]))

    ls = [interpolate(p) for p in ls]

    for line in ls:
        may.plot3d(line[:, 0], line[:, 1], line[:, 2],
                   color=colour, tube_radius=tube_radius)

def plot_cell_vispy(lines, boundary=None, clf=True, colours=None,
                    randomise_colours=True, tube_radius=1., **kwargs):
    if clf:
        clear_vispy_canvas()
    
    if colours is None:
        hues = n.linspace(0, 1, len(lines) + 1)[:-1]
        colours = [hsv_to_rgb(hue, 1, 1) for hue in hues]
        if randomise_colours:
            random.shuffle(colours)
    elif not isinstance(colours, (list, tuple)):
        colours = [colours for _ in lines]
    i = 0
    segments = []
    segment_colours = []
    segment_radii = []
    if not isinstance(tube_radius, list):
        tube_radius = [tube_radius for _ in range(len(lines))]
    for (line, colour, radius) in zip(lines, colours, tube_radius):
        vprint('Plotting line {} / {}\r'.format(i, len(lines)-1),
               False)
        i += 1
        if len(colour) == 4 and colour[-1] == 0:
            continue
        for segment in line:
            if len(segment) < 4:
                continue
            segments.append(segment)
            segment_colours.append(colour)
            segment_radii.append(radius)
            # plot_line_vispy(segment,
            #                 clf=False, colour=colour, **kwargs)
    plot_lines_vispy(segments, colours=segment_colours,
                     tube_radius=segment_radii, clf=clf, **kwargs)
    
    if boundary is not None:
        draw_bounding_box_vispy(boundary, tube_radius=tube_radius[0])

                
def draw_bounding_box_vispy(shape, colour=(0, 0, 0), tube_radius=1.):
    if shape is not None:
        if isinstance(shape, (float, int)):
            shape = ensure_shape_tuple(shape)
        if len(shape) == 3:
            shape = (0, shape[0], 0, shape[1], 0, shape[2])

    xmin, xmax, ymin, ymax, zmin, zmax = shape
    ls = []
    ls.append(n.array([[xmax, ymax, zmin],[xmax, ymax, zmax]]))
    ls.append(n.array([[xmax, ymin, zmin],[xmax, ymin, zmax]]))
    ls.append(n.array([[xmin, ymax, zmin],[xmin, ymax, zmax]]))
    ls.append(n.array([[xmin, ymin, zmin],[xmin, ymin, zmax]]))
    ls.append(n.array([[xmin, ymax, zmax],[xmax, ymax, zmax]]))
    ls.append(n.array([[xmin, ymin, zmax],[xmax, ymin, zmax]]))
    ls.append(n.array([[xmin, ymax, zmin],[xmax, ymax, zmin]]))
    ls.append(n.array([[xmin, ymin, zmin],[xmax, ymin, zmin]]))
    ls.append(n.array([[xmax, ymin, zmax],[xmax, ymax, zmax]]))
    ls.append(n.array([[xmin, ymin, zmax],[xmin, ymax, zmax]]))
    ls.append(n.array([[xmax, ymin, zmin],[xmax, ymax, zmin]]))
    ls.append(n.array([[xmin, ymin, zmin],[xmin, ymax, zmin]]))

    ls = [interpolate(p) for p in ls]

    # plot_lines_vispy(ls, colours=['black' for _ in ls],
    #                  tube_radius=tube_radius, zero_centroid=False)
    plot_lines_vispy(ls, clf=False, colours=[colour for _ in ls],
                     zero_centroid=False, tube_radius=tube_radius)
    # for line in ls:
    #     plot_line_vispy(line, clf=False, colour=colour,
    #                     zero_centroid=False, tube_radius=tube_radius)

    global vispy_canvas
    vispy_canvas.central_widget.children[0].camera.center = (
        shape[1] / 2., shape[1] / 2., shape[1] / 2.)
    vispy_canvas.update()

def interpolate(p, num=10):
    p1, p2 = p
    return n.array([n.linspace(i, j, num) for i, j in zip(p1, p2)]).T

def cell_to_povray(filen, lines, shape):
    from jinja2 import Environment, FileSystemLoader
    env = Environment(loader=FileSystemLoader(
        '/home/asandy/devel/pyknotid/pyknotid/templates'))
    template = env.get_template('cell.pov')

    colours = n.linspace(0, 1, len(lines) + 1)[:-1]
    colours = n.array([hsv_to_rgb(c, 1, 1) for c in colours])

    coloured_segments = []
    for line, colour in zip(lines, colours):
        for segment in line:
            if len(segment) > 3:
                coloured_segments.append((segment, colour))

    with open(filen, 'w') as fileh:
        fileh.write(template.render(lines=coloured_segments))
    

def vispy_save_png(filename):
    img = vispy_canvas.render()
    import vispy.io as io
    io.write_png(filename, img)

def plot_sphere_mollweide_vispy(func, circle_points=50, depth=2,
                                edge_color=None,
                                cmap='hsv',
                                smooth=0,
                                mesh='circles',
                                **kwargs):
    '''func must be a function of sphere angles theta, phi'''
    
    # from vispy.scene import Sphere, ArcballCamera
    from vispy.scene import TurntableCamera, Mesh

    if mesh == 'circles':
        vertices, indices = circles_ellipse_mesh(circle_points, depth)
    else:
        vertices, indices = recursive_ellipse_mesh(circle_points, depth)
    vertices[:, 0] *= 2*n.sqrt(2)
    vertices[:, 1] *= n.sqrt(2)
    mesh = Mesh(vertices, indices)

    md = mesh._meshdata
    vertices = md.get_vertices()

    values = n.zeros(len(vertices))

    print('pre')
    thetas = []
    phis = []
    for i, vertex in enumerate(vertices):
        if i % 10 == 0:
            vprint('\ri = {} / {}'.format(i, len(vertices)), newline=False)

        intermediate = n.arcsin(vertex[1] / n.sqrt(2))
        theta = n.arcsin((2*intermediate + n.sin(2*intermediate)) / n.pi)
        phi = n.pi * vertex[0] / (2*n.sqrt(2) * n.cos(intermediate))

        # theta = n.arccos(vertex[2])
        # phi = n.arctan2(vertex[1], vertex[0])

        if n.isnan(theta):
            theta = 0.0
            print('theta', vertex)
        if n.isnan(phi):
            phi = 0.0
            print('phi', vertex)

        thetas.append(theta)
        phis.append(phi)
        values[i] = func(theta + n.pi/2, phi + n.pi)
    vprint()

    print('thetas', n.min(thetas), n.max(thetas))
    print('phis', n.min(phis), n.max(phis))

    colours = n.zeros((len(values), 4))
    max_val = n.max(values)
    min_val = n.min(values)
    unique_values = n.unique(colours)
    max_val += (1. + 1./len(unique_values))*(max_val - min_val)
    diff = (max_val - min_val)

    import matplotlib.pyplot as plt
    cm = plt.get_cmap(cmap)
    for i in range(len(colours)):
        colours[i] = cm(((values[i] - min_val) / diff))

    colours[:, -1] = 1.

    faces = md.get_faces()
    for si in range(smooth):
        new_colours = [[n.array(row) for _ in range(3)]
                       for row in colours]
        for i, face in enumerate(faces):
            new_colours[face[0]].append(colours[face[1]])
            new_colours[face[0]].append(colours[face[2]])
            new_colours[face[1]].append(colours[face[0]])
            new_colours[face[1]].append(colours[face[2]])
            new_colours[face[2]].append(colours[face[0]])
            new_colours[face[2]].append(colours[face[1]])

        new_colours = n.array([n.average(cs, axis=0) for cs in new_colours])

        colours = new_colours

    md.set_vertex_colors(colours)

    ensure_vispy_canvas()
    clear_vispy_canvas()
    vispy_canvas.view.camera = TurntableCamera(fov=0, elevation=90.)
    vispy_canvas.view.add(mesh)
    vispy_canvas.show()

def circles_ellipse_mesh(radial=5, azimuthal=100):
    angles = n.linspace(0, 2*n.pi, azimuthal + 1)[:-1]
    radii = n.linspace(0, 1., radial)

    offsets = n.zeros(len(angles))
    offsets[::2] += 2*n.pi / azimuthal

    vertices = []
    for i, radius in enumerate(radii):
        cur_angles = angles
        if i % 2 == 0:
            cur_angles += 0.5* (2*n.pi) / azimuthal
        points = n.zeros((len(angles), 3))
        points[:, 0] = n.cos(angles) * radius
        points[:, 1] = n.sin(angles) * radius


        vertices.append(points)

    vertices = n.vstack(vertices)

    indices = []
    num_angles = len(angles)
    for num, radius in enumerate(radii[:-1]):
        base_index = num_angles * num
        next_num = num + 1
        for i in range(len(angles)):
            cur_index = base_index + i
            next_index = base_index + (i + 1) % num_angles
            next_r_index_1 = base_index + ((i - 1) % num_angles) + num_angles
            next_r_index_2 = base_index + i + num_angles
            indices.append((cur_index, next_r_index_1, next_r_index_2))
            indices.append((cur_index, next_r_index_2, next_index))

    return vertices, n.array(indices)
                            
def plot_vertices_indices(vertices, indices):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots()

    ax.plot(vertices[:, 0], vertices[:, 1], 'o')

    for triangle in indices:
        ax.plot([vertices[triangle[0]][0],
                 vertices[triangle[1]][0],
                 vertices[triangle[2]][0]],
                [vertices[triangle[0]][1],
                 vertices[triangle[1]][1],
                 vertices[triangle[2]][1]])

    fig.show()
    return fig, ax
                          


def recursive_ellipse_mesh(init_num_points=10, depth=2):
    triangle = n.array([[0., 0.],
                        [1., 0.],
                        [0.5, 1.5]])

    angles = n.linspace(0, 2*n.pi, init_num_points + 1)[:-1]
    xs = n.sin(angles)
    ys = n.cos(angles)
    centre = n.array([0., 0.])
        
    vertices = []
    indices = []
    vertices_dict = {}

    for i in range(len(angles)):
        triangle = n.array([centre, (xs[i], ys[i]),
                            (xs[(i+1) % len(angles)],
                             ys[(i+1) % len(angles)])])
        indices.extend(get_subtriangles(triangle, vertices,
                                        vertices_dict, 0,
                                        depth))
                                                        

    # indices = get_subtriangles(triangle, vertices, {}, 0, depth)

    return (n.array(vertices), n.array(indices))
                         
    

def get_subtriangles(points, vertices, vertices_dict, depth, max_depth):
    if depth > max_depth:
        indices = []
        for point in points:
            point = tuple(point)
            if tuple(point) not in vertices_dict:
                vertices.append(point)
                vertices_dict[point] = len(vertices) - 1
            indices.append(vertices_dict[point])
                
        return [indices]

    arr_points = n.array(points)

    centre = n.average(arr_points, axis=0)
    half_edges = arr_points + 0.5*(n.roll(arr_points, -1, axis=0) - arr_points)

    new_points = n.array([arr_points[0],
                          half_edges[0],
                          arr_points[1],
                          half_edges[1],
                          arr_points[2],
                          half_edges[2]])

    subtriangles = []
    for i in range(6):
        new_triangle = n.array([centre, new_points[i], new_points[(i+1) % 6]])
        subtriangles.extend(get_subtriangles(new_triangle, vertices,
                                             vertices_dict,
                                             depth+1, max_depth))
    return subtriangles


    
def plot_sphere_lambert_vispy(func, circle_points=50, depth=2,
                              edge_color=None,
                              cmap='hsv',
                              smooth=0,
                              mesh='circles',
                              **kwargs):
    '''func must be a function of sphere angles theta, phi'''
    
    # from vispy.scene import sphere, arcballcamera
    from vispy.scene import TurntableCamera, Mesh

    if mesh == 'circles':
        vertices, indices = circles_ellipse_mesh(circle_points, depth)
    else:
        vertices, indices = recursive_ellipse_mesh(circle_points, depth)
    # vertices[:, 0] *= 2*n.sqrt(2)
    # vertices[:, 1] *= n.sqrt(2)
    vertices[:, 0] *= 2
    vertices[:, 1] *= 2
    mesh = Mesh(vertices, indices)

    md = mesh._meshdata
    vertices = md.get_vertices()

    values = n.zeros(len(vertices))

    print('pre')
    thetas = []
    phis = []
    for i, vertex in enumerate(vertices):
        if i % 10 == 0:
            vprint('\ri = {} / {}'.format(i, len(vertices)), newline=False)

        # print('vertex is', vertex)
        # print('radius is', np.sqrt(vertex[0]**2 + vertex[1]**2))
        # print('angle is', np.arctan2(vertex[1], vertex[0]))

        # intermediate = n.arcsin(vertex[1] / n.sqrt(2))
        # theta = n.arcsin((2*intermediate + n.sin(2*intermediate)) / n.pi)
        # phi = n.pi * vertex[0] / (2*n.sqrt(2) * n.cos(intermediate))

        theta = 2*np.arccos(np.sqrt(vertex[0]**2 + vertex[1]**2)/2.)
        phi = np.arctan2(vertex[1], vertex[0])

        # theta = n.arccos(vertex[2])
        # phi = n.arctan2(vertex[1], vertex[0])

        if n.isnan(theta):
            theta = 0.0
            print('theta', vertex)
        if n.isnan(phi):
            phi = 0.0
            print('phi', vertex)

        thetas.append(theta)
        phis.append(phi)
        values[i] = func(theta + n.pi/2, phi + n.pi)
    vprint()

    print('thetas', n.min(thetas), n.max(thetas))
    print('phis', n.min(phis), n.max(phis))

    colours = n.zeros((len(values), 4))
    max_val = n.max(values)
    min_val = n.min(values)
    unique_values = n.unique(colours)
    max_val += (1. + 1./len(unique_values))*(max_val - min_val)
    diff = (max_val - min_val)

    import matplotlib.pyplot as plt
    cm = plt.get_cmap(cmap)
    for i in range(len(colours)):
        colours[i] = cm(((values[i] - min_val) / diff))

    colours[:, -1] = 1.

    faces = md.get_faces()
    for si in range(smooth):
        new_colours = [[n.array(row) for _ in range(3)]
                       for row in colours]
        for i, face in enumerate(faces):
            new_colours[face[0]].append(colours[face[1]])
            new_colours[face[0]].append(colours[face[2]])
            new_colours[face[1]].append(colours[face[0]])
            new_colours[face[1]].append(colours[face[2]])
            new_colours[face[2]].append(colours[face[0]])
            new_colours[face[2]].append(colours[face[1]])

        new_colours = n.array([n.average(cs, axis=0) for cs in new_colours])

        colours = new_colours

    md.set_vertex_colors(colours)

    ensure_vispy_canvas()
    clear_vispy_canvas()
    vispy_canvas.view.camera = TurntableCamera(fov=0, elevation=90.)
    vispy_canvas.view.add(mesh)
    vispy_canvas.show()


def plot_sphere_lambert_sharp_vispy(func, circle_points=50, depth=2,
                                    output_size=500,
                                    edge_color=None,
                                    cmap='brg',
                                    smooth=0,
                                    mesh='circles',
                                    **kwargs):
    '''func must be a function of sphere angles theta, phi'''
    
    from vispy.scene import TurntableCamera, Mesh

    if mesh == 'circles':
        vertices, indices = circles_ellipse_mesh(circle_points, depth)
    else:
        vertices, indices = recursive_ellipse_mesh(circle_points, depth)
    vertices[:, 0] *= 2
    vertices[:, 1] *= 2
    mesh = Mesh(vertices, indices)

    md = mesh._meshdata
    vertices = md.get_vertices()

    values = n.zeros(len(vertices))

    print('pre')
    thetas = []
    phis = []
    for i, vertex in enumerate(vertices):
        if i % 10 == 0:
            vprint('\ri = {} / {}'.format(i, len(vertices)), newline=False)

        theta = 2*np.arccos(np.sqrt(vertex[0]**2 + vertex[1]**2)/2.)
        phi = np.arctan2(vertex[1], vertex[0])

        if n.isnan(theta):
            theta = 0.0
            print('theta', vertex)
        if n.isnan(phi):
            phi = 0.0
            print('phi', vertex)

        thetas.append(theta)
        phis.append(phi)
        values[i] = func(theta + n.pi/2, phi + n.pi)
    vprint()

    return thetas, phis, values

    import svgwrite as svg
    d = svg.Drawing()

    import matplotlib.pyplot as plt
    cmap = plt.get_cmap(cmap)
    
    max_value = n.max(values)
    min_value = n.min(values)
    def normalise(v):
        return (v - min_value) / (max_value - min_value)
    for tri_i, triangle in enumerate(indices):
        v1 = vertices[triangle[0]]
        v2 = vertices[triangle[1]]
        v3 = vertices[triangle[2]]

        c1 = values[triangle[0]]
        c2 = values[triangle[1]]
        c3 = values[triangle[2]]

        offset_x = 2.001412
        offset_y = 2.0214
        v1_x = (v1[0] + offset_x) / 4. * output_size
        v1_y = (v1[1] + offset_y) / 4. * output_size

        v2_x = (v2[0] + offset_x) / 4. * output_size
        v2_y = (v2[1] + offset_y) / 4. * output_size

        v3_x = (v3[0] + offset_x) / 4. * output_size
        v3_y = (v3[1] + offset_y) / 4. * output_size

        c_x = n.average([v1_x, v2_x, v3_x])
        c_y = n.average([v1_y, v2_y, v3_y])

        v12_x = (v1_x + v2_x) / 2.
        v12_y = (v1_y + v2_y) / 2.
        v13_x = (v1_x + v3_x) / 2.
        v13_y = (v1_y + v3_y) / 2.
        v23_x = (v2_x + v3_x) / 2.
        v23_y = (v2_y + v3_y) / 2.

        cmap_1 = cmap(normalise(c1))
        rgb_cmap1 = [int(c*255) for c in cmap_1[:3]]
        d.add(d.polygon([[v1_x, v1_y],
                         [v12_x, v12_y],
                         [c_x, c_y],
                         [v13_x, v13_y]],
                        fill='rgb({},{},{})'.format(*rgb_cmap1),
                        stroke='rgb({},{},{})'.format(*rgb_cmap1),
                        stroke_width='0.5'))

        cmap_2 = cmap(normalise(c2))
        rgb_cmap2 = [int(c*255) for c in cmap_2[:3]]
        d.add(d.polygon([[v2_x, v2_y],
                         [v12_x, v12_y],
                         [c_x, c_y],
                         [v23_x, v23_y]],
                        fill='rgb({},{},{})'.format(*rgb_cmap2),
                        stroke='rgb({},{},{})'.format(*rgb_cmap2),
                        stroke_width='0.5'))

        cmap_3 = cmap(normalise(c3))
        rgb_cmap3 = [int(c*255) for c in cmap_3[:3]]
        d.add(d.polygon([[v3_x, v3_y],
                         [v13_x, v13_y],
                         [c_x, c_y],
                         [v23_x, v23_y]],
                        fill='rgb({},{},{})'.format(*rgb_cmap3),
                        stroke='rgb({},{},{})'.format(*rgb_cmap3),
                        stroke_width='0.5'))

        # print('cols', c1, cmap_1, c2, cmap_2, c3, cmap_3)

    return d

        
                    
        
        


def get_coloured_subtriangles(points, point_values, vertices, values, vertices_dict, depth, max_depth):
    if depth > max_depth:
        indices = []
        for point, value in zip(points, point_values):
            point = tuple(point)
            if tuple(point) not in vertices_dict:
                vertices.append(point)
                values.append(value)
                vertices_dict[point] = len(vertices) - 1
            indices.append(vertices_dict[point])
                
        return [indices]

    arr_points = n.array(points)

    centre = n.average(arr_points, axis=0)
    half_edges = arr_points + 0.5*(n.roll(arr_points, -1, axis=0) - arr_points)

    new_points = n.array([arr_points[0],
                          half_edges[0],
                          arr_points[1],
                          half_edges[1],
                          arr_points[2],
                          half_edges[2]])

    subtriangles = []
    for i in range(6):
        new_triangle = n.array([centre, new_points[i], new_points[(i+1) % 6]])
        subtriangles.extend(get_subtriangles(new_triangle, vertices,
                                             vertices_dict,
                                             depth+1, max_depth))
    return subtriangles


def point_inside_triangle(point, triangle):
    px, py = point
    p1x, p1y = triangle[0]
    p2x, p2y = triangle[1]
    p3x, p3y = triangle[2]
    alpha = (((p2y - p3y)*(px - p3x) + (p3x - p2x)*(py - p3y)) /
             ((p2y - p3y)*(p1x - p3x) + (p3x - p2x)*(p1y - p3y)))
    if alpha < 0:
        return False
    beta = (((p3y - p1y)*(px - p3x) + (p1x - p3x)*(py - p3y)) /
            ((p2y - p3y)*(p1x - p3x) + (p3x - p2x)*(p1y - p3y)))
    if beta < 0:
        return False
    gamma = 1.0 - alpha - beta
    if gamma < 0:
        return False

    return True



def PointInsideTriangle2(pt,tri):
    '''checks if point pt(2) is inside triangle tri(3x2). @Developer

    source: https://stackoverflow.com/a/20949123/2469283
    '''
    a = 1/(-tri[1,1]*tri[2,0]+tri[0,1]*(-tri[1,0]+tri[2,0])+ \
        tri[0,0]*(tri[1,1]-tri[2,1])+tri[1,0]*tri[2,1])
    s = a*(tri[2,0]*tri[0,1]-tri[0,0]*tri[2,1]+(tri[2,1]-tri[0,1])*pt[0]+ \
        (tri[0,0]-tri[2,0])*pt[1])
    if s<0: return False
    else: t = a*(tri[0,0]*tri[1,1]-tri[1,0]*tri[0,1]+(tri[0,1]-tri[1,1])*pt[0]+ \
              (tri[1,0]-tri[0,0])*pt[1])
    return ((t>0) and (1-s-t>0))



'''
Database download module
========================

To download the database, call :func:`download_database`.

The other functions in this module provide basic functionality for
checking where the database is stored, and deleting old versions if
necessary.

API documentation
-----------------

'''

from functools import wraps
from os.path import realpath, dirname, exists, join, abspath
from os import mkdir, remove

def find_database(db_version=None):
    '''Returns the path to the knots.db file.

    find_db looks in the following locations, in order of precedence:

    1. The local folder (containing getdb.py). This is convenient if
       you have built your own database.
    2. The directory returned by appdirs.user_data_dir (depends on the OS).

    If the database cannot be found, an exception is raised.

    You can download a prebuilt database using :func:`download_database`.

    Parameters
    ----------
    db_version : int
        The database version to find. Defaults to None, in which case the
        current db_version from :mod:`pyknotid.catalogue.database` is used.
    '''
    local_filen = join(dirname(__file__),
                       'knots.db')
    if exists(local_filen):
        return local_filen

    if db_version is None:
        from pyknotid.catalogue import db_version
    local_filen = join(dirname(realpath(__file__)),
                       'knots_{}.db'.format(db_version))
    if exists(local_filen):
        return local_filen

    import appdirs
    app_dir = appdirs.user_data_dir('pyknotid')
    app_dir_filen = join(app_dir, 'knots_{}.db'.format(db_version))
    if exists(app_dir_filen):
        return app_dir_filen

    raise IOError('Could not find a knots database file. You can '
                  'download one using '
                  '`pyknotid.catalogue.download_database()`.')


def download_target_dir():
    '''Returns the directory to which the knots database will be
    downloaded.'''
    import appdirs
    return appdirs.user_data_dir('pyknotid')

def download_database():
    '''Downloads the knots database to :func:`download_target_dir`.
    '''
    dirn = download_target_dir()
    if not exists(dirn):
        mkdir(dirn)

    from pyknotid.catalogue import db_version
    db_name = 'knots_{}.db'.format(db_version)
    filen = join(dirn, db_name)
    if exists(filen):
        raise IOError('A file named {} already exists.'.format(filen))

    import requests
    from tqdm import tqdm
    # import shutil
    r = requests.get('https://github.com/SPOCKnots/pyknotid/releases/download/init/{}'.format(db_name),
                     stream=True)

    total_size = int(r.headers.get('content-length', 0))

    with open(filen, 'wb') as f:
        with tqdm(total=total_size, unit='B', unit_scale=True) as pbar:
            for data in r.iter_content(32*1024):
                pbar.update(32*1024)
                f.write(data)

    print('Successfully downloaded the new database file. Run '
          'pyknotid.catalogue.getdb.clean_old_databases to delete '
          'old database versions.')

def clean_old_databases():
    '''Deletes old database files (all but the most recent version).'''

    dirn = download_target_dir()
    import glob
    filens = glob.glob(join(dirn, 'knots_*.db'))
    versions = sorted(filens, key=lambda j: int(j.split('_')[1][:-3]))

    print('Found databases: {}'.format(', '.join(versions)))
    for version in versions[:-1]:
        print('Deleting {}'.format(version))
        remove(version)

def clean_all_databases():
    '''Deletes all database files.'''

    dirn = download_target_dir()
    import glob
    filens = glob.glob(join(dirn, 'knots_*.db'))
    versions = sorted(filens, key=lambda j: int(j.split('_')[1][:-3]))

    print('Found databases: {}'.format(', '.join(versions)))
    for version in versions:
        print('Deleting {}'.format(version))
        remove(version)


def require_database(func):
    '''Decorator that causes a function to query find_database before
    returning.'''
    @wraps(func)
    def new_func(*args, **kwargs):
        find_database()
        return func(*args, **kwargs)
    return new_func



'''.. image:: random_walk_length_30.png
   :scale: 50%
   :alt: A closed random walks with 30 steps
   :align: center

This module contains classes and functions for working with knots and
links as three-dimensional space curves, or calling functions
elsewhere in pyknotid to perform topological analysis. Functionality
includes manipulating knots/links via translation, rotation and
scaling, plotting diagrams, finding crossings and identifying knots.

Different knot classes
----------------------

pyknotid includes the following classes for topological calculation:

- :doc:`spacecurve`: Provides functions for calculations on a single
  curve, including plotting, some geometrical properties and finding
  crossings in projection.

- :doc:`knot`: Provides functions for topological calculations on a
  single curve, such as the Alexander polynomial or Vassiliev
  invariants.

- :doc:`openknot`: Provides functions for topological calculations on
  an open curve that does not form a closed loop. Open curves are
  topologically trivial from a mathematical perspective, but can be
  analysed in terms of the topology of different closures.

- :doc:`link`: Provides the same interface to collections of multiple
  curves, and can calculate linking invariants.

- :doc:`periodiccell`: Provides some convenience functions for
  managing collections of curves in periodic boundaries.

Creating space curves
---------------------

The space curve classes are specified via N by 3 arrays of points in
three dimensions, representing a piecewise linear curve.

For instance, the following code produces and plots a
:class:`~pyknotid.spacecurves.knot.Knot` from a set of manually
specified points::

  import numpy as np
  from pyknotid.spacecurves import Knot

  points = np.array([[9.0, 0.0, 0.0],
                     [0.781, 4.43, 2.6],
                     [-4.23, 1.54, -2.6],
                     [-4.5, -7.79, -7.35e-16],
                     [3.45, -2.89, 2.6],
                     [3.45, 2.89, -2.6],
                     [-4.5, 7.79, 0.0],
                     [-4.23, -1.54, 2.6],
                     [0.781, -4.43, -2.6]])

  k = Knot(points)
  k.plot()

.. image:: trefoil_few_points.png
   :align: center
   :alt: A trefoil knot specified by vertex points
   :scale: 50%

The :doc:`pyknotid.make module<../make/index>` provides functions for
creating many types of example knots, such as torus knots or some
specific knot types::

  import numpy as np
  from pyknotid.make import torus_knot

   k = torus_knot(7, 4)
   k.plot()


.. image:: p7_q4__torus_knot.png
   :align: center
   :scale: 50%
   :alt: A p=7 and q=4 knot produced by the above code

'''

from pyknotid.spacecurves.spacecurve import SpaceCurve
from pyknotid.spacecurves.knot import Knot
from pyknotid.spacecurves.link import Link
from pyknotid.spacecurves.openknot import OpenKnot
from pyknotid.spacecurves.periodiccell import Cell

__all__ = ('SpaceCurve', 'Knot', 'Link', 'OpenKnot', 'Cell', )




'''
Utility functions
=================
'''

import sys
import numpy as n

def vprint(string='', newline=True, condition=True):
    '''
    Verbose print; prints the given string if the input condition
    is met.

    Pyknot2 uses this to make it easy to display changing counters.

    Parameters
    ==========
    string : str
        The string to print.
    newline : bool
        Whether to automatically append a newline.
    condition : bool
        The condition that must be met for the print to occur.
    '''
    if not condition:
        return
    sys.stdout.write(string)
    if newline:
        sys.stdout.write('\n')
    sys.stdout.flush()

def mag(v):
    '''
    Returns the magnitude of the vector v.

    Parameters
    ----------
    v : ndarray
        A vector of any dimension.
    '''
    return n.sqrt(v.dot(v))

def get_rotation_matrix(angles):
    '''
    Returns a rotation matrix based on sequentially rotating
    around the 3 given axes (i.e. by multiplication of the component
    rotation matrices).

    Parameters
    ----------
    angles : iterable
        The psi, theta and phi angles
    '''
    
    phi, theta, psi = angles
    sin = n.sin
    cos = n.cos
    rotmat = n.array([
            [cos(theta)*cos(psi),
             -1*cos(phi)*sin(psi) + sin(phi)*sin(theta)*cos(psi),
             sin(phi)*sin(psi) + cos(phi)*sin(theta)*cos(psi)],
            [cos(theta)*sin(psi),
             cos(phi)*cos(psi) + sin(phi)*sin(theta)*sin(psi),
             -1*sin(phi)*cos(psi) + cos(phi)*sin(theta)*sin(psi)],
            [-1*sin(theta), sin(phi)*cos(theta), cos(phi)*cos(theta)]])
    return rotmat


def ensure_shape_tuple(shape):
    '''If the input is a number, returns a tuple with that number
    repeated three times, otherwise just returns the input.
    '''
    if isinstance(shape, (float, long, int)):
        return (shape, shape, shape)
    return shape



'''
Space curve geometry
====================

Functions for evaluating geometrical properties of space curves.
'''

import numpy as n

def arclength(points, include_closure=True):
    '''
    Returns the arclength as the sum of lengths of each segment
    in the piecewise linear line.

    Parameters
    ----------
    points : array-like
        Nx3 array of points in the line
    include_closure : bool
        Whether to include the distance between the final and
        first points. Defaults to True.
    '''

    if len(points) < 2:
        return 0.
        
    lengths = n.roll(points, -1, axis=0) - points
    length_mags = n.sqrt(n.sum(lengths*lengths, axis=1))
    arclength = n.sum(length_mags[:-1])
    if include_closure:
        arclength += length_mags[-1]
    return arclength
    

def radius_of_gyration(points):
    '''
    Returns the radius of gyration of the given vertices, assuming
    each has equal weight (and ignoring the connecting lines).

    Parameters
    ----------
    points : array-like
        Nx3 array of points in the line.
    '''
    av_pos = n.average(points, axis=0)
    diffs = (points - av_pos)**2
    rogs = n.sum(diffs, 1)
    rog = n.average(rogs)
    return n.sqrt(rog)


# def persistences(points, step=None):
#     '''
#     Returns a set of xs and ys for persistence length calculation.
#     '''

#     if step is None:
#         import pyknot.
#         step = n.average



'''Functions for converting between representations, e.g. from sympy
polynomials to the json lists stored in the database models.  '''

import sympy as sym
import numpy as n    
import json

## 1) Conversion from database to sympy

def sympify_alexander(coeffs):
    '''Takes an Alexander polynomial as a list of coefficients, and
    returns a sympy polynomial representation.'''
    t = sym.var('t')
    index = 0
    result = 0
    for coeff in coeffs:
        result += coeff * t**index
        index += 1
    return result

def sympify_jones(coeffs):
    '''Takes a Jones polynomial as a list of (coefficient, index) tuples,
    and returns a sympy polynomial representation.'''
    q = sym.var('q')
    result = 0
    for coeff, index in coeffs:
        result += coeff*q**index
    return result

def sympify_homfly(coeffs):
    '''Takes a HOMFLY polynomial as a list of (coefficient, a_index,
    z_index) tuples, and returns a sympy polynomial representation.
    '''
    a = sym.var('a')
    z = sym.var('z')
    result = 0
    for coeff, a_index, z_index in coeffs:
        result += coeff * a**a_index * z**z_index
    return result

def json_to_alexander(s):
    '''Takes an Alexander polynomial json serialisation and returns the
    sympy polynomial.'''
    return sympify_alexander(json.loads(s))

def json_to_jones(s):
    '''Takes an Jones polynomial json serialisation and returns the
    sympy polynomial.'''
    return sympify_jones(json.loads(s))

def json_to_homfly(s):
    '''Takes a HOMFLY polynomial json serialisation and returns the
    sympy polynomial.'''
    return sympify_homfly(json.loads(s))
    

## 2) Conversion from sympy to database json

def desympify_alexander(p):
    """Takes a sympy Alexander polynomial p, and returns a tuple of
    coefficients in order.
    """
    if p == 1:
        return (1, )
    p = p.expand()
    terms = p.as_terms()[0]
    coeffs = n.zeros((len(terms),2),dtype=int)
    for i in range(len(terms)):
        entry = terms[i]
        coeffs[i,0] = int(entry[1][0][0])
        coeffs[i,1] = int(entry[1][1][0])

    coeffs = coeffs[n.argsort(coeffs[:,1])]
    if coeffs[0][0] < 0:
        coeffs[:,0] *= -1

    cmin = coeffs[0,1]
    cmax = coeffs[-1,1]

    cs = n.zeros(cmax-cmin+1,dtype=int)
    for entry in coeffs:
        cs[entry[1]-cmin] = int(entry[0])

    return tuple(map(int,cs))

def desympify_jones(p):
    """Takes a sympy Jones polynomial p, and returns a tuple of
    (coefficient, index) tuples.
    """
    if p == 1:
        return (1, )
    p = p.expand()
    terms = p.as_terms()[0]
    coeffs = n.zeros((len(terms), 2), dtype=int)
    for i in range(len(terms)):
        entry = terms[i]
        coeffs[i,0] = int(entry[1][0][0])
        coeffs[i,1] = int(entry[1][1][0])

    coeffs = coeffs[n.argsort(coeffs[:, 1])]
    # if coeffs[0][0] < 0:
    #     coeffs[:, 0] *= -1

    cmin = coeffs[0, 1]
    cmax = coeffs[-1, 1]

    cs = []
    for entry in coeffs:
        cs.append((int(entry[0]), int(entry[1])))
        # Need int to convert from numpy.int64

    return tuple(cs)

def desympify_homfly(p):
    """Takes a sympy HOMFLY polynomial p, and returns a tuple of
    (coefficient, index_a, index_z) tuples.
    """
    if p == 1:
        return (1, )
    p = p.expand()
    terms = p.as_terms()[0]
    coeffs = n.zeros((len(terms), 3), dtype=int)
    for i in range(len(terms)):
        entry = terms[i]
        coeffs[i,0] = int(entry[1][0][0])
        coeffs[i,1] = int(entry[1][1][0])
        coeffs[i,2] = int(entry[1][1][1])

    inds = n.lexsort((coeffs[:,2], coeffs[:,1]))
    coeffs = coeffs[inds]

    cmin = coeffs[0, 1]
    cmax = coeffs[-1, 1]

    cs = []
    for entry in coeffs:
        cs.append((int(entry[0]), int(entry[1]), int(entry[2])))
        # Need int to convert from numpy.int64

    return tuple(cs)

def alexander_to_json(p):
    '''Takes an Alexander polynomial, and returns a json serialised list
    of its coefficients.'''
    return json.dumps(desympify_alexander(p))

def jones_to_json(p):
    '''Takes a Jones polynomial, and returns a json serialised list
    of its coefficients.'''
    return json.dumps(desympify_jones(p))

def homfly_to_json(p):
    '''Takes a HOMFLY polynomial, and returns a json serialised list of
    its coefficients.'''
    return json.dumps(desympify_homfly(p))

## 3) Conversion from rdf to sympy

def rdf_poly_to_sympy(p, vars=None):
    '''Converts polynomials as stored in knot atlas rdf data to sympy
    objects. The process is *very* crude (trim off <math> xml, replace
    characters with python equivalents), but seems to suffice for the
    data currently harvested. 

    vars should be a list of variables in the polynomial
    (e.g. ['t','q']). Appropriate sympy variables are constructed to
    represent these.

    The conversion is done using eval and exec, so you should be *very
    sure* that your rdf doesn't contain malicious code!

    '''
    p = p.strip()
    p = p.replace('<math>', '')
    p = p.replace('</math>', '')
    p = p.strip()
    p = p.replace('{', '(')
    p = p.replace('}', ')')
    p = p.replace('^', '**')
    p = p.replace(' + ','+')
    p = p.replace(' +','+')
    p = p.replace('+ ','+')
    p = p.replace(' - ','-')
    p = p.replace(' -','-')
    p = p.replace('- ','-')
    p = p.replace(' ','*')

    for entry in vars:
        exec(''.join((entry, ' = sym.var(\'', entry, '\')')))

    return eval(p)


## 4) User interface convenience functions (standard names)

def db2py_alexander(s):
    return json_to_alexander(s)

def py2db_alexander(p):
    return alexander_to_json(p)

def db2py_jones(s):
    return json_to_jones(s)

def py2db_jones(p):
    return jones_to_json(p)

def db2py_homfly(s):
    return json_to_homfly(s)

def py2db_homfly(p):
    return homfly_to_json(p)

## 5) Converters between invariants

def homfly_to_jones(p):
    '''Takes a homfly polynomial p, and returns the Jones polynomial
    through variable substitution.'''
    a = sym.var('a')
    z = sym.var('z')
    q = sym.var('q')

    np = p.subs(a, 1/q)
    np = np.subs(z, sym.sqrt(q) - 1/sym.sqrt(q))
    np = np.expand()
    return np

def homfly_other_chirality(p):
    '''Takes a homfly polynomial p, and replaces a with 1/a to give the
    polynomial of the same knot with opposite chirality.'''
    if p == 1:
        return 1
    a = sym.var('a')
    np = p.subs(a, 1/a)
    np = np.expand()
    return np

def jones_other_chirality(p):
    '''Takes a jones polynomial p, and replaces q with 1/q to give the
    polynomial of the same knot with opposite chirality.'''
    if p == 1:
        return 1
    q = sym.var('q')
    np = p.subs(q, 1/q)
    np = np.expand()
    return np



'''Database module
---------------

pyknotid looks up knot information via a prebuilt sqlite database,
accessed using the peewee ORM. Other ORMs are not currently supported.

The model class is :class:`pyknotid.catalogue.database.Knot`,
documented below. For generic documentation about using the database,
see :doc:`index`.

pyknotid also includes functions for creating a database from scratch
(using knot information from the Knot Atlas), and improving an
existing database by calculating new invariants or pulling information
from other sources such as the KnotInfo database. These functions can
be found in ``pyknotid.catalogue.build`` and
``pyknotid.catalogue.improve``, which are not included in the indexed
documentation here.

API documentation
~~~~~~~~~~~~~~~~~

'''

from peewee import *
import json
import os
from os.path import realpath, dirname, exists, join
from pyknotid.catalogue import converters

from pyknotid.catalogue.getdb import find_database

if not ('READTHEDOCS' in os.environ and os.environ['READTHEDOCS'] == 'True'):
    DB_LOCATION = find_database()
    db = SqliteDatabase(DB_LOCATION)
    db.connect()
else:
    DB_LOCATION = None
    db = None

class BaseModel(Model):
    class Meta(object):
        database = db


# If updating the db structure, don't forget to update
# pyknotid.catalogue.db_version!
class Knot(BaseModel):
    '''Peewee model for storing a knot in a database.'''

    name = CharField(null=True)
    '''The actual name (if any), e.g. trefoil'''

    identifier = CharField(null=True)
    '''The standard knot notation, e.g. 3_1 for trefoil'''

    min_crossings = IntegerField(null=True)
    '''Minimal crossing number for the knot, e.g. 3 for trefoil'''

    signature = IntegerField(null=True)
    '''The knot signature'''

    determinant = IntegerField(null=True)
    '''The knot determinant (Alexander polynomial at -1)'''

    alexander_imag_3 = IntegerField(null=True)
    '''The absolute value of the Alexander polynomial at
    exp(2 pi I / 3). This will always be an integer.'''

    alexander_imag_4 = IntegerField(null=True)
    '''The absolute value of the Alexander polynomial at
    exp(2 pi I / 4) == I. This will always be an integer.'''

    alexander = TextField(null=True)
    '''Alexander polynomial, stored as a json list of coefficients from
    lowest to highest index, including zeros if there are any jumps in
    index.
    '''

    jones = TextField(null=True)
    '''Jones polynomial, stored as a json list of coefficients and indices
    for each monomial.'''

    homfly = TextField(null=True)
    '''HOMFLY-PT polynomial, stored as a json list.'''

    unknotting_number = IntegerField(null=True)
    '''Unknotting number, stored as an integer.'''

    hyperbolic_volume = CharField(null=True)
    '''Hyperbolic volume, stored as a string to avoid precision
    problems.'''

    conway_notation = CharField(null=True)
    '''Conway notation, as a string.'''

    gauss_code = CharField(null=True)
    '''Gauss code, as a string.'''

    planar_diagram = CharField(null=True)
    '''Planar diagram representation, as a string.'''

    dt_code = CharField(null=True)
    '''Dowker-Thistlethwaite code, as a string.'''

    two_bridge = CharField(null=True)
    '''Two-bridge notation, as a string.'''

    fibered = BooleanField(null=True)
    '''Whether the knot is fibered or not.'''

    vassiliev_2 = IntegerField(null=True)
    '''The Vassiliev invariant of order 2.'''

    vassiliev_3 = IntegerField(null=True)
    '''The Vassiliev invariant of order 3.'''

    symmetry = CharField(null=True, max_length=25)
    '''The symmetry type of the knot; reversible,
    positive amphichiral, negative amphichiral fully
    amphichiral or chiral.'''

    planar_writhe = IntegerField(null=True)
    '''The writhe of the minimal diagram described by the DT_code. This is
    not necessarily unique (see Perko pair, I think?).'''

    def __str__(self):
        if self.name:
            return '<Knot {} ({})>'.format(self.identifier, self.name)
        return '<Knot {}>'.format(self.identifier)

    def __repr__(self):
        return str(self)

    def pretty_print(self):
        '''Pretty print all information contained about self.'''
        if self.name:
            print('Name: {}'.format(self.name))
        if self.identifier:
            print('Identifier: {}'.format(self.identifier))
        if self.min_crossings:
            print('Min crossings: {}'.format(self.min_crossings))
        if self.fibered is not None:
            print('Fibered: {}'.format(self.fibered))
        if self.gauss_code:
            print('Gauss code: {}'.format(self.gauss_code))
        if self.planar_diagram:
            print('Planar diagram: {}'.format(self.planar_diagram))
        if self.dt_code:
            print('DT code: {}'.format(self.dt_code))
        if self.determinant:
            print('Determinant: {}'.format(self.determinant))
        if self.signature:
            print('Signature: {}'.format(self.signature))
        if self.alexander:
            print('Alexander: {}'.format(
                converters.json_to_alexander(self.alexander)))
        if self.jones:
            print('Jones: {}'.format(
                converters.json_to_jones(self.jones)))
        if self.homfly:
            print('HOMFLY: {}'.format(
                converters.json_to_homfly(self.homfly)))
        if self.hyperbolic_volume:
            print('Hyperbolic volume: {}'.format(self.hyperbolic_volume))
        if self.vassiliev_2 is not None:
            print('Vassiliev order 2: {}'.format(self.vassiliev_2))
        if self.vassiliev_3 is not None:
            print('Vassiliev order 3: {}'.format(self.vassiliev_3))
        if self.symmetry is not None:
            print('Symmetry: {}'.format(self.symmetry))

    def space_curve(self, verbose=True, **kwargs):
        '''Returns a Knot object representing this knot.'''

        if self.dt_code is None:
            raise ValueError('No DT code is known, cannot create '
                             'space curve.')

        from pyknotid.representations import DTNotation
        import numpy as np
        d = DTNotation(self.dt_code)
        k = d.representation(verbose=verbose, **kwargs).space_curve(
            verbose=verbose)
        k.points += 0.0001 * np.random.random(k.points.shape)
        k.close()
        return k

    def url(self):
        '''The guessed url of this knot in the Knot Atlas. This page may not
        actually exist or be populated.

        '''
        return 'http://katlas.org/wiki/{}'.format(self.identifier)

    def open_url(self):
        import webbrowser
        webbrowser.open(self.url())

    def retrieve_image_path(self, **kwargs):
        images_folder = join(
            dirname(realpath(__file__)), 'diagrams_with_mirrors')

        name = self.identifier

        filen = _name_to_filen(name)

        filen = join(images_folder, filen)

        if not exists(filen):
            return None

        return filen


def _name_to_filen(name):
    if name.startswith('K'):
        name = name[1:]
        if int(name[:2]) == 11:
            digits = 3
        else:
            digits = 4
        filen = '{{}}_{{:0{}}}.png'.format(digits).format(name[:3], int(name[3:]))

    else:
        filen = name + '.png'

    return filen




'''
Identify module
---------------

Functions for identifying knots based on their name or invariants.

API documentation
~~~~~~~~~~~~~~~~~
'''

# Most imports are placed in from_invariants, so that the module can work
# even if no database is available.

from pyknotid.catalogue.getdb import require_database


@require_database
def get_knot(name):
    '''Returns from the database the Knot with the given identifier.

    For instance, :code:`trefoil = get_knot('3_1')`.
    '''
    k = first_from_invariants(id=name)
    if k is None:
        raise ValueError('No knot with this name was found')
    return k


@require_database
def first_from_invariants(**kwargs):
    '''Returns the first Knot by crossing number (and arbitrary
    ordering within that) with the given invariant conditions.

    Parameters
    ----------
    **kwargs :
        Any set of invariant conditions. The accepted arguments are
        the same as for :func:`from_invariants`.
    '''
    return from_invariants(return_query=True, **kwargs).first()


@require_database
def from_invariants(return_query=False, **kwargs):
    '''Takes invariants as kwargs, and does the appropriate conversion to
    return a list of database objects matching all the given criteria.

    .. note:: This only searches within the indexed database
              available.  Some invariant options return only results where the
              invariant both matches *and* is known, others return those that
              match *or* are not known. Check the source if depending on
              accurate results.

    Does *not* support all available invariants. Currently, searching
    is supported by:

    Parameters
    ----------
    identifier or name or id : str
        The name of the knot following knot atlas conventions, e.g. '3_1'
    min_crossings : int
        The minimal crossing number of the knot.
    max_crossings : int
        The maximal known crossing number of the knot. This may be higher
        than its actual crossing number, it serves only to prune the
        results list.
    signature : int
        The signature invariant.
    unknotting_number : int
        The unknotting number of the knot.
    alexander or alex : sympy
        The Alexander polynomial, provided as a sympy expression in a
        single variable (ideally 't').
    determinant or alexander_imag_2: int
        The Alexander polynomial at -1 (== exp(Pi I))
    alexander_imag_3 : int
        The abs of the Alexander polynomial at exp(2 Pi I / 3)
    alexander_imag_4 : int
        The abs of the Alexander polynomial at exp(Pi I / 2)
    roots : iterable
        The abs of the Alexander polnomial at the given roots, assumed
        to start at 2, e.g. passing (3, 2, 1) is the same as identifying
        at determinant=3, alexander_imag_3=2, alexander_imag_4=1. An
        entry of None means the value is ignored in the lookup.
    jones : sympy
        The Jones polynomial, provided as a sympy expression in a single
        variable (ideally 'q').
    homfly : sympy
        The HOMFLY-PT polynomial, provided as a sympy expression in two
        variables.
    hyperbolic_volume or hyp_vol or hypvol : float or str
        The hyperbolic volume of the knot complement. The lookup is a
        string comparison based on the given number of significant digits.
    vassiliev_order_2 or vassiliev_2 or v_2 or v2 : int
        The Vassiliev invariant of order 2. This will not prune knots
        where this invariant is not known (specifically, those with 12 or
        more crossings).
    vassiliev_order_3 or vassiliev_3 or v_3 or v3 : int
        The Vassiliev invariant of order 3. This will not prune knots
        where this invariant is not known (specifically, those with 12 or
        more crossings).
    symmetry : string
        The symmetry of the knot, one of 'reversible',
        'positive amphicheiral', 'negative amphicheiral', 'chiral'.
    writhe or planar_writhe : int
        The writhe of the knot's minimal diagram as recorded by its dt
        code. This is not necessarily unique, only the value
        of the dt code stored is given.
    other : iterable
        A list of other peewee terms that can be chained in ``where()``
        calls, e.g. ``database.Knot.min_crossings < 5``. This provides
        more flexibility than the other options.
    return_query : bool
        If True, returns the database iterator for the objects, otherwise
        returns a list. Defaults to False (i.e. the list). This will
        be much slower if the list is very large, but is convenient
        for most searches.

    '''

    from . import database as db
    from pyknotid.catalogue.database import Knot
    from pyknotid.catalogue import converters

    db.db.get_conn()

    _root_to_attr = {2: Knot.determinant,
                     3: Knot.alexander_imag_3,
                     4: Knot.alexander_imag_4}

    conditions = []
    for invariant, value in kwargs.items():
        invariant = invariant.lower()
        if invariant in ['name', 'id', 'identifier']:
            conditions.append(Knot.identifier == value)
        elif invariant == 'min_crossings':
            conditions.append(Knot.min_crossings == value)
        elif invariant == 'max_crossings':
            conditions.append(Knot.min_crossings <= value)
        elif invariant == 'signature':
            conditions.append(Knot.signature == value)
        elif invariant in ['determinant', 'alexander_imag_2',
                           'alex_imag_2', 'weak_2']:
            conditions.append(Knot.determinant == value)
        elif invariant in ['alexander_imag_3', 'alex_imag_3', 'weak_3']:
            conditions.append(Knot.alexander_imag_3 == value)
        elif invariant in ['alexander_imag_4', 'alex_imag_4', 'weak_4']:
            conditions.append(Knot.alexander_imag_4 == value)
        elif invariant in ['roots']:
            for root, result in zip(range(2, len(value)+2), value):
                if result is not None:
                    conditions.append(_root_to_attr[root] == result)
        elif invariant == 'unknotting_number':
            conditions.append(Knot.unknotting_number == value)
        elif invariant in ['alexander', 'alex']:
            val = converters.py2db_alexander(value)
            conditions.append(Knot.alexander == val)
        elif invariant == 'jones':
            val = converters.py2db_jones(value)
            chiral_val = converters.py2db_jones(
                converters.jones_other_chirality(value))
            conditions.append(Knot.jones << [val, chiral_val])
        elif invariant in ['homfly', ]:
            val = converters.py2db_homfly(value)
            chiral_val = converters.py2db_homfly(
                converters.homfly_other_chirality(value))
            conditions.append(Knot.homfly << [val, chiral_val])
        elif invariant in ['hypvol', 'hyp_vol', 'hyperbolic_volume']:
            conditions.append(Knot.hyperbolic_volume % '{}*'.format(str(value)))
        elif invariant in ['vassiliev_order_2', 'vassiliev_2', 'v_2', 'v2']:
            conditions.append((Knot.vassiliev_2 == value) |
                              (Knot.vassiliev_2 >> None))
        elif invariant in ['vassiliev_order_3', 'vassiliev_3', 'v_3', 'v3']:
            conditions.append((Knot.vassiliev_3 == value) |
                              (Knot.vassiliev_3 == -1*value) |
                              (Knot.vassiliev_3 >> None))
        elif invariant == 'symmetry':
            conditions.append(Knot.symmetry == value.lower())
        elif invariant in ('writhe', 'planar_writhe'):
            conditions.append(Knot.planar_writhe == value)
        elif invariant == 'other':
            for condition in value:
                conditions.append(condition)

    selection = Knot.select()
    for condition in conditions:
        selection = selection.where(condition)
    selection = selection.order_by(Knot.min_crossings)
    if return_query:
        return selection

    selection = list(selection)

    return sorted(selection, key=_sort_func)


def _sort_func(k):
    ident = k.identifier

    if ident.startswith('K'):
        ident = ident[1:]
        if 'a' in ident:
            parts = ident.split('a')
            jump = 0
        elif 'n' in ident:
            parts = ident.split('n')
            jump = 1e-7
        crossings = int(parts[0])
        number = int(parts[1])
    else:
        parts = ident.split('_')
        crossings = int(parts[0])
        number = int(parts[1])
        jump = 0

    return crossings + number * 1e-15 + jump



import numpy as n


def get_rotation_angles(number):
    '''
    Returns a list of theta, phi values, approximately evenly
    distributed on the sphere.

    Uses the generalised spiral points algorithm explained in
    E B Saff and
    A B J Kuijlaars, Distributing many points on a sphere, The
    Mathematical Intelligencer 19(1) 1997.
    '''

    angles = n.zeros((number, 2))
    angles[0] = n.array([n.arccos(-1), 0])

    for k in range(2, number+1):
        h_k = -1. + (2. * (k - 1)) / (number - 1)
        theta = n.arccos(h_k)
        phi = (angles[k-2, 1] + 3.6/n.sqrt(number) *
               1. / n.sqrt(1 - h_k**2)) % (2*n.pi)
        angles[k-1, 0] = theta
        angles[k-1, 1] = phi
    angles[-1, 1] = 0.  # Last phi will be inf otherwise

    return angles

def rotate_vector_to_top(vector):
    '''Returns a rotation matrix that will rotate the given vector to
    point upwards.

    Parameters
    ----------
    vector : ndarray
        The (3D) vector to rotate.
    '''

    theta = n.arccos(vector[2] / n.linalg.norm(vector))
    phi = n.arctan2(vector[1], vector[0])
    return rotate_to_top(theta, phi)

def rotate_to_top(theta, phi):

    '''
    Returns a rotation matrix that will rotate a sphere such that
    the given positions are at the top.

    Parameters
    ----------
    theta : float
        The latitudinal variable.
    phi : float
        The longitudinal variable.
    '''

    chi = -1 * phi
    alpha = theta

    cc = n.cos(chi)
    sc = n.sin(chi)
    ca = n.cos(alpha)
    sa = n.sin(alpha)
    first_rotation = n.array([[cc, -sc, 0],
                              [sc, cc, 0],
                              [0, 0, 1]])
    second_rotation = n.array([[ca, 0, -sa],
                               [0, 1, 0],
                               [sa, 0, ca]])

    return second_rotation.dot(first_rotation)

def rotate_axis_angle(axis, angle):
    axis = n.array(axis) / n.linalg.norm(axis)
    ux, uy, uz = axis

    ct = n.cos(angle)
    st = n.sin(angle)
    
    arr = n.array(
        [[ct + ux**2*(1-ct), ux*uy*(1-ct) - uz*st, ux*uz*(1-ct) + uy*st],
         [uy*ux*(1-ct) + uz*st, ct + uy**2*(1-ct), uy*ux*(1-ct) - ux*st],
         [uz*ux*(1-ct) - uy*st, uz*uy*(1-ct) + ux*st, ct + uz**2*(1-ct)]])

    return arr
    



'''
PeriodicCell
------------

Tools for working with a periodic cell of spacecurves.

API documentation
~~~~~~~~~~~~~~~~~
'''
from pyknotid.utils import ensure_shape_tuple
from pyknotid.spacecurves import Knot, OpenKnot
import numpy as n
import numpy as np

class Cell(object):
    '''Class for holding the vertices of some number of lines with
    periodic boundary conditions.

    Parameters
    ==========
    lines : list
        Must be a list of Knots or ndarrays of vertices.
    shape : tuble or int
        The shape of the cell, in whatever units the lines use.
    periodic : bool
        Whether the cell is periodic. If True, lines are marked as
        'nth' or 'loop' in self.line_types. Defaults to True.
    '''

    def __init__(self, lines, shape, periodic=True, cram=False, downsample=None):
        self.shape = ensure_shape_tuple(shape)

        lines = [l for l in lines if len(l) > 1]
        if downsample is not None:
            lines = [l[::downsample] for l in lines]
        lines = list(map(_interpret_line, lines))
        if cram:
            lines = [_cram_into_cell(l, self.shape) for l in lines]
        self.lines = [_cut_line_at_jumps(l, self.shape) for l in lines]
        self.periodic = periodic

        self.line_types = None
        if periodic:
            self.line_types = [_test_periodicity(l, self.shape) for l in self.lines]

    @classmethod
    def from_qwer(cls, qwer, shape, **kwargs):
        '''Returns an instance of Cell from a quartet of differently
        classified lines in periodic boundaries.

        Parameters
        ----------
        qwer : tuple
            Should be a 4-tuple of lists q, w, e, r. q is closed
            loops, w is lines with non-trivial homology, e is lines that
            terminate on the boundaries of the cell, r is any remaining
            (unclassified) lines.
        shape : int or tuple
            The size of the cell along each axis. If a single number
            is passed, all axes are assumed to be the same length.

        '''
        q, w, e, r = qwer
        if len(w) > 0 and isinstance(w[0], tuple):
            w = [l[2] for l in w]
        if len(e) > 0 and isinstance(e[0], tuple):
            e = [l[2] for l in e]
        output = cls(q+w+e, shape, **kwargs)
        output.qwer = (q, w, e, r)
        return output

    def append(self, line, cram=False):
        line = _interpret_line(line)
        if cram:
            line = _cram_into_cell(line)
        self.lines.append(line)

    def plot(self, boundary=True, clf=True, tube_radius=0.5,
             length_colours=None,
             **kwargs):
        from pyknotid.visualise import plot_cell
        boundary = self.shape if boundary else None

        if length_colours is not None:
            if 'colours' in kwargs:
                raise ValueError(
                    'colours and length_colours cannot both be set')

            q, w, e, r = self.qwer
            from pyknotid.spacecurves.openknot import OpenKnot

            lengths = []
            for line in q + w + e:
                k = OpenKnot.from_periodic_line(line, self.shape)
                length = k.arclength()
                lengths.append(length)
            total_length = np.sum(lengths)
            lengths = [l / total_length for l in lengths]

            colours = []
            for length in lengths:
                for bound, colour in length_colours[:-1]:
                    if length < bound:
                        colours.append(colour)
                        break
                else:
                    colours.append(length_colours[-1])
            kwargs['colours'] = colours

        plot_cell(self.lines, boundary=boundary, clf=clf,
                  tube_radius=tube_radius, **kwargs)

    def smooth(self, repeats=1, window_len=10):
        '''Smooth each line in the curve, equivalent to
        :meth:`~pyknotid.spacecurves.spacecurve.SpaceCurve.smooth`.

        '''
        from pyknotid.spacecurves import Knot
        new_lines = []
        for i, line in enumerate(self.lines):
            new_segments = []
            for segment in line:
                if len(segment) > window_len:
                    k = Knot(segment)
                    k.smooth(repeats, periodic=False, window_len=window_len)
                    new_segments.append(k.points)
                else:
                    new_segments.append(segment)
            new_lines.append(new_segments)
        self.lines = new_lines

    def to_povray(self, filen, spline='cubic_spline'):
        from pyknotid.visualise import cell_to_povray
        cell_to_povray(filen, self.lines, self.shape)

    def get_lengths(self):
        from pyknotid.spacecurves.openknot import OpenKnot
        lengths = []
        for line in self.lines:
            points = n.vstack(line)
            k = OpenKnot.from_periodic_line(points, self.shape,
                                            perturb=False)
            lengths.append(k.arclength())
        return lengths

    def simplify(self, num=1, cut_selection='uniform', **kwargs):
        from pyknotid.simplify import octree

        for i in range(num):
            print('i = {} / {}'.format(i, num))
            ot = octree.OctreeCell.from_cell_lines(self.lines, self.shape,
                                                cut_selection=cut_selection,
                                                **kwargs)
            ot.simplify()

            self.lines = [_cut_line_at_jumps(l, self.shape) for l in
                        ot.get_lines()]

    def linking_matrix(self):
        '''Get the linking numbers of each line in the cell with every
        other.
        '''
        from pyknotid.spacecurves.openknot import OpenKnot
        from pyknotid.spacecurves.link import Link
        from collections import defaultdict

        lines = self.lines
        
        types = ['loop' if _is_closed_loop(
            OpenKnot.from_periodic_line(np.vstack(l), self.shape,
                                        perturb=False).points)
                 else 'line' for l in lines]

        linkings = np.zeros((len(lines), len(lines)))

        linkings = {}

        for i1, details in enumerate(zip(lines, types)):
            line1, line_type1 = details
            print('i1 is', i1, len(lines))
            for i2, other_details in enumerate(zip(lines[i1:], types[i1:])):
                i2 += i1
                print('i2 is', i2)
                line2, line_type2 = other_details

                if line_type1 == 'loop' and line_type2 == 'loop':

                    linking = get_linking_between_loops(
                        line1, line2, self.shape, same_loop=(i1==i2))
                    
                    print('linking is', linking)

                    if linking:
                        linkings[(i1, i2)] = linking

                elif line_type1 == 'loop' and line_type2 == 'line':
                    linking = get_linking_between_loop_and_line(
                        line1, line2, self.shape)

                # Same as previous condition, just the opposite order
                elif line_type1 == 'line' and line_type2 == 'loop':
                    linking = get_linking_between_loop_and_line(
                        line2, line1, self.shape)

                elif line_type1 == 'line' and line_type2 == 'line':
                    pass

        return linkings
        

    # def efficient_linking_matrix(self):
    #     '''Get the linking numbers of each line in the cell with every
    #     other.
    #     '''
    #     # Should this be the periodic linking number or not?

    #     linkings = np.zeros((len(self.lines),
    #                          len(self.lines)))

    #     import chelpers

    #     all_crossings = []

    #     for i1, line in enumerate(self.lines):
    #         segments_1 = line

    #         cum_lengths_1 = np.cumsum(map(len, segments_1))
    #         print('i1 is', i1, len(self.lines))

    #         for i2, other_line in enumerate(self.lines[i1+1:]):
    #             i2 += 1
    #             segments_2 = line
    #             cum_lengths_2 = np.cumsum(map(len, segments_2))

    #             steps_2 = [
    #                 np.roll(segment[:, :2], -1, axis=0) - segment[:, :2] for
    #                 segment in segments_1]
    #             step_lengths_2 = []
    #             for steps in steps_2:
    #                 step_lengths_2.append(np.array(
    #                     [np.sqrt(np.sum(row**2))
    #                      for row in steps]))
    #             max_step_lengths_2 = [np.max(ls) for ls in
    #                                       step_lengths_2]

    #             current_linking = 0
    #             current_offset = (0, 0, 0)

    #             crossings = []

    #             for segment_1, current_index in zip(segments_1, cum_lengths_1):
    #                 for i, point in enumerate(segment_1[:-1]):
    #                     next_point = segment_1[i+1]
    #                     crossing_index = current_index + i
    #                     dv = next_point - point
    #                     for crossing_index_2, segment_2, lengths_2, max_step_length in zip(
    #                             cum_lengths_2, segments_2,
    #                             step_lengths_2,
    #                             max_step_lengths_2):

    #                         new_crossings = chelpers.find_crossings(
    #                             point, dv,
    #                             segment_2,
    #                             lengths_2,
    #                             crossing_index,
    #                             0,
    #                             max_step_length
    #                             )
    #                         if new_crossings:
    #                             new_crossings = np.array(new_crossings)
    #                             new_crossings[::2, 1] += crossing_index_2
    #                             new_crossings[1::2, 0] += crossing_index_2
    #                             # new_crossings[:, 1] += crossing_index_2
    #                             crossings.append(new_crossings)

    #             if crossings:
    #                 all_crossings.append((i1, i2, np.vstack(crossings)))

    #     linkings = []

    #     return all_crossings
                    
        
def _interpret_line(line):
    if isinstance(line, Knot):
        return line.points
    elif isinstance(line, n.ndarray):
        return line

    return ValueError('Lines must be Knots or ndarrays.')

def _test_periodicity(line, shape):
    closing_vector = (line[-1][-1] - line[0][0]) / n.array(shape)
    if n.any(closing_vector > 0.2):
        return 'nth'
    return 'loop'

def _cut_line_at_jumps(line, shape):
    x, y, z = shape
    shape = n.array(shape)
    if x < 0 or y < 0 or z < 0:
        return line
    line = line.copy()
    i = 0
    out = []
    while i < (len(line)-1):
        cur = line[i]
        nex = line[i+1]
        if n.any(n.abs(nex-cur) > 0.9*shape):
            first_half = line[:(i+1)]
            second_half = line[(i+1):]
            out.append(first_half)
            line = second_half
            i = 0
        else:
            i += 1
    out.append(line)
    return out

def _cram_into_cell(line, shape):
    '''Imposes the shape as periodic boundary conditions.'''
    shape = n.array(shape)
    dx, dy, dz = shape

    points = n.array(line).copy()
    for i in range(1, len(points)):
        prev = points[i-1]
        cur = points[i]
        rest = points[i:]
        if cur[0] < 0:
            rest[:, 0] += dx
        if cur[0] > dx:
            rest[:, 0] -= dx
        if cur[1] < 0:
            rest[:, 1] += dy
        if cur[1] > dy:
            rest[:, 1] -= dy
        if cur[2] < 0:
            rest[:, 2] += dz
        if cur[2] > dz:
            rest[:, 2] -= dz

    return points

def _is_closed_loop(line, cutoff=5.):
    closing_distance = np.sqrt(np.sum((line[-1] - line[0])**2))
    return closing_distance < cutoff


from pyknotid.spacecurves.link import Link

class BoundingBox(object):
    def __init__(self, l):
        if hasattr(l, 'points'):
            # crudely get the points if l is a
            # Knot or SpaceCurve
            l = l.points

        self.mins = np.min(l, axis=0)
        self.maxs = np.max(l, axis=0)

    def intersects(self, b):
        b1 = self
        b2 = b

        b1xmax, b1ymax, b1zmax = b1.maxs
        b1xmin, b1ymin, b1zmin = b1.mins

        b2xmax, b2ymax, b2zmax = b2.maxs
        b2xmin, b2ymin, b2zmin = b2.mins

        return (b1xmax > b2xmin and
                b1xmin < b2xmax and
                b1ymax > b2ymin and
                b1ymin < b2ymax and
                b1zmax > b2zmin and
                b1zmin < b2zmax)

    def translations_to(self, b, shape):
        '''
        Returns the translations of b1 that could make it overlap b2.
        '''
        b1 = self
        b2 = b

        size1 = b1.maxs - b1.mins
        size2 = b2.maxs - b2.mins

        if not isinstance(shape, (int, float)):
            assert len(set(shape)) == 1
            shape = float(shape[0])

        # should these have +1 and -1?
        steps_mins = np.floor((b2.mins - b1.maxs) / shape).astype(np.int) + 1
        steps_maxs = np.floor((b2.maxs - b1.mins) / shape).astype(np.int)

        return steps_mins, steps_maxs

        distance_from_b2_to_b1 = b2.mins - b1.mins
        num_steps_from_b2_to_b1 = distance_from_b2_to_b1 / shape


def get_linking_between_loops(line1, line2, shape,
                              same_loop=False,
                              periodic=True):
    if periodic:
        l = Link.from_periodic_lines((np.vstack(line1),
                                      np.vstack(line2)),
                                     shape,
                                     perturb=False)
    else:
        l = Link((line1, line2))
    b1 = BoundingBox(l.lines[0].points)
    b2 = BoundingBox(l.lines[1].points)

    # Get the range of x, y and z translations under
    # which the bounding boxes collide

    translations = b2.translations_to(b1, shape)

    print('translations are', translations)

    linkings = {}

    for dx in range(translations[0][0], translations[1][0] + 1):
        for dy in range(translations[0][1], translations[1][1] + 1):
            for dz in range(translations[0][2], translations[1][2] + 1):
                if same_loop:
                    if dx == dy == dz == 0:
                        continue
                print('dx dy dz', dx, dy, dz, translations[0], translations[1])
                points1 = l.lines[0].points.copy()
                points2 = l.lines[1].points.copy()

                points2 += np.array([dx, dy, dz]) * np.array(shape)

                b2 = BoundingBox(points2)

                if not b1.intersects(b2):
                    continue

                l_new = Link((points1, points2), verbose=False)
                l_new.rotate()

                linking_number = l_new.linking_number()
                if linking_number:
                    linkings[(dx, dy, dz)] = linking_number

    return linkings

def get_linking_between_loop_and_line(loop, line, shape,
                                      periodic=True):

    loop= np.vstack(loop)
    line = np.vstack(line)
    loop = Knot.from_periodic_line(loop, shape)
    line = OpenKnot.from_periodic_line(line, shape)

    line_closure = (line.points[-1] - line.points[0]) / shape[0]
    line_closure = np.round(line_closure).astype(np.int)

    print('closure is', line_closure)

    b1 = BoundingBox(line.points)
    b2 = BoundingBox(loop.points)

    # Get the range of x, y and z translations under
    # which the bounding boxes collide

    translations = b2.translations_to(b1, shape)

    print('translations are', translations)

    linkings = {}

    from pyknotid.spacecurves.link import Link
    from pyknotid.spacecurves.rotation import rotate_vector_to_top
    from pyknotid.utils import get_rotation_matrix

    random_matrix = get_rotation_matrix(np.random.random(size=3) * 2 * np.pi)

    # loop._apply_matrix(rotation_matrix)
    # line._apply_matrix(rotation_matrix)

    loop._apply_matrix(random_matrix)
    line._apply_matrix(random_matrix)

    to_top_matrix = rotate_vector_to_top(line.points[-1] - line.points[0])

    loop._apply_matrix(to_top_matrix)
    line._apply_matrix(to_top_matrix)

    # Switch x and z axes
    # loop.points[:, 0], loop.points[:, 2] = loop.points[:, 2], loop.points[:, 0]
    # line.points[:, 0], line.points[:, 2] = line.points[:, 2], line.points[:, 0]

    axis_rotation = get_rotation_matrix((np.pi / 2., 0, 0))

    line._apply_matrix(axis_rotation)
    loop._apply_matrix(axis_rotation)

    for dx in range(translations[0][0], translations[1][0] + 1):
        for dy in range(translations[0][1], translations[1][1] + 1):
            for dz in range(translations[0][2], translations[1][2] + 1):

                print(dx, dy, dz)

                translation = random_matrix.dot(np.array((dx, dy, dz)) * np.array(shape))
                translation = to_top_matrix.dot(translation)
                translation = axis_rotation.dot(translation)

                print('translation is', translation)

                points1 = line.points.copy()
                points2 = loop.points.copy()
                points1[-1] = points1[0]

                points2 += translation

                b1 = BoundingBox(points1)
                b2 = BoundingBox(points2)


                if not b1.intersects(b2):
                    continue

                l = Link((points1, points2), verbose=False)

                linking_number = l.linking_number(include_closures=False)

                # if linking_number:
                #     import ipdb
                #     ipdb.set_trace()


                if linking_number:
                    linkings[(dx, dy, dz)] = linking_number
                
    return linkings
                

                # We can calculate the linking number as a link in the current plane, as the crossings will be made up later

    # 0.1) Choose a random rotation

    # 1) Rotate the line to have homology vector in z=0
    # 2) For every translation where the loop overlaps the line at all...
    #    ...but factoring out translations that are multiples of the line's homology vector!
    # 3) Extend the line as far as necessary
    # 4) Calculate the linking

    # For every position where the 


loop1 = np.array([[0., 0, 0],
                  [5, 0, 0],
                  [5, 5, 0],
                  [0, 5, 0],
                  [0, 0.1, 0]])

loop2 = np.array([[2.5, 2.5, -3],
                  [2.5, 2.5, 3],
                  [7, 8, 3],
                  [7, 8, -3],
                  [2.7, 2.65, -3]])

loop3 = loop2 + np.array([30., 0, 0])



'''
IO
==

Functions for saving and loading knot data from files.
'''

import json
import numpy as n


def to_json_file(points, filen):
    '''
    Writes the given points to the given filename, using
    json format.

    Parameters
    ----------
    points : array-like
        nx3 array of points to save
    filen : str
        The (relative) filename to save in
    '''
    points = n.array(points).tolist()
    with open(filen, 'w') as fileh:
        json.dump(points, fileh)

def from_json_file(filen):
    '''
    Loads an array of points from the given filename assuming
    json format, and returns the result as a numpy array.

    Parameters
    ----------
    filen : str
        The (relative) filename to load from
    '''
    with open(filen, 'r') as fileh:
        points = json.load(fileh)

    return n.array(points)


def from_csv(filen, index_col=None, header_row=None, **kwargs):
    '''
    Loads an array of points from the given filename parsed
    as a csv.
    
    .. note:: This function requires pandas to be installed.

    .. note:: For data in a space-separated column format,
              pass the argument `sep=' '` to read in as a csv.

    Parameters
    ----------
    filen : str
        The (relative) filename to load from.
    index_col: int or None
        The column that indexes the rows, defaults to None
        meaning no such column is present.
    header_row: int or None
        The row with header information (i.e. column names), with
        all previous rows ignored. Defaults to None, which means
        no such header exists.

    **kwargs :
        Passed directly to pandas.DataFrame.from_csv.
    '''
    import pandas as pn
    df = pn.DataFrame.from_csv(filen, index_col=index_col,
                               header_row=header_row, **kwargs)
    return df.as_matrix()



